import logging
import time
from abc import ABC, abstractmethod
from typing import List, Optional

from steamship import Block, Steamship
from steamship.agents.schema import AgentContext, Metadata
from steamship.agents.tools.audio_transcription.assembly_speech_to_text_tool import (
    AssemblySpeechToTextTool,
)


class Transport(ABC):
    client = Steamship
    """Experimental base class to encapsulate a communication channel

    Intended use is:

        class MyBot(PackageService):
            transport: Transport

            def __init__(self):
                self.transport = TelegramTransport(bot_token=self.config.telegram_bot_token)

            def instance_init(self):
                webhook_url = self.context.invocable_url + 'respond'
                self.transport.instance_init(webhook_url=webhook_url)

            def respond(self):
                self.transport.send(blocks=[Block(text="Got it!")], chat_id="abc")

    Note that this experimental sketch is just a first draft. Among its quirks:

    - The idea of re-using Blocks as the medium of message format, which aligns chatting with the rest of our code.
      - For example, to send back an image, one would do so via a block.

    - It doesn't yet try to model inbound messages or chat_ids. That's an encapsulation leak left for future exploration.

    """

    def __init__(self, client):
        self.client = client

    def instance_init(self, *args, **kwargs):
        logging.info(f"Transport initializing: {self.__class__.__name__}")
        start = time.time()
        self._instance_init(*args, **kwargs)
        end = time.time()
        logging.info(f"Transport initialized in {end - start} seconds: {self.__class__.__name__}")

    @abstractmethod
    def _instance_init(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def _instance_deinit(self, *args, **kwargs):
        raise NotImplementedError

    def instance_deinit(self, *args, **kwargs):
        logging.info(f"Transport deinitializing: {self.__class__.__name__}")
        start = time.time()
        self._instance_deinit(*args, **kwargs)
        end = time.time()
        logging.info(f"Transport deinitialized in {end - start} seconds: {self.__class__.__name__}")

    def send(self, blocks: List[Block], metadata: Optional[Metadata] = None):
        metadata = metadata or {}
        if blocks is None or len(blocks) == 0:
            logging.info(f"Skipping send of 0 blocks: {self.__class__.__name__}")
            return

        logging.info(f"Sending {len(blocks)} blocks: {self.__class__.__name__}")
        start = time.time()
        self._send(blocks, metadata)
        end = time.time()
        logging.info(
            f"Sending {len(blocks)} blocks in {end - start} seconds: {self.__class__.__name__}"
        )

    @abstractmethod
    def _send(self, blocks: List[Block], metadata: Metadata):
        raise NotImplementedError

    def parse_inbound(self, payload: dict, context: Optional[dict] = None) -> Optional[Block]:
        message = self._parse_inbound(payload, context)

        if message.url and not message.text:
            context = AgentContext()
            context.client = self.client
            transcriptions = AssemblySpeechToTextTool(
                blockifier_plugin_config={
                    "enable_audio_intelligence": False,
                    "speaker_detection": False,
                }
            ).run([Block(text=message.url)], context=context)
            message.text = transcriptions[0].text
        return message

    @abstractmethod
    def _parse_inbound(self, payload: dict, context: Optional[dict] = None) -> Optional[Block]:
        raise NotImplementedError

    def info(self) -> dict:
        logging.info(f"Getting transport info: {self.__class__.__name__}")
        start = time.time()
        info = self._info()
        end = time.time()
        logging.info(f"Transport info fetched in {end - start} seconds: {self.__class__.__name__}")
        return info

    @abstractmethod
    def _info(self) -> dict:
        raise NotImplementedError



import os
from enum import Enum

from steamship.base.configuration import Configuration
from steamship.base.error import SteamshipError


class RuntimeEnvironments(str, Enum):
    REPLIT = "replit"
    LOCALHOST = "localhost"


def _interactively_get_key(env: RuntimeEnvironments):
    print(
        """Get your free API key here: https://steamship.com/account/api

You'll get immediate access to our SDK for AI models, including OpenAI, GPT, Cohere, and more.
"""
    )

    api_key = input("Paste your API key to run: ")

    while len(api_key.strip()) == 0:
        api_key = input("API Key: ")

    os.environ["STEAMSHIP_API_KEY"] = api_key

    if env == RuntimeEnvironments.REPLIT:
        print(
            """
This key is set temporarily. In the future, you can:
- Set the STEAMSHIP_API_KEY Replit Secret
- Close and re-open any Replit shells to make sure secrets are refreshed.

"""
        )
    elif env == RuntimeEnvironments.LOCALHOST:
        print(
            """
This key is set temporarily. In the future, you can:
- Set the STEAMSHIP_API_KEY environment variable
- Run `ship login` to create a ~/.steamship.json credential file

"""
        )


def _report_error_and_exit(env: RuntimeEnvironments):
    if env == RuntimeEnvironments.REPLIT:
        print(
            """To run this Replit, you will need a Steamship API Key.

1) If you're viewing someone else's Replit, clone it

2) Visit https://steamship.com/account/api to get a key

3) Add your key as a Replit secret named STEAMSHIP_API_KEY

4) Close and re-open any shells to make sure your new secret is available

Then try running again!"""
        )
    elif env == RuntimeEnvironments.LOCALHOST:
        print(
            """To run this script, you will need a Steamship API Key.

1) Visit https://steamship.com/account/api to get a key

2) Set your key as the environment variable STEAMSHIP_API_KEY

Then try running again!

If you have pip-installed `steamship`, you can also try setting your key by simply running `ship login`.
"""
        )
    exit(-1)


def check_environment(env: RuntimeEnvironments, interactively_set_key: bool = True):
    # This will try loading from STEAMSHIP_API_KEY and also ~/.steamship.json
    try:
        config = Configuration()

        # If an API key is set, we're good to go!
        if config.api_key:
            return
    except SteamshipError:
        # The Configuration object will throw an error if there is no API Key found.
        # Since that error is expected from the context of this function, we pass on it to handle it in a more
        # user-interactive way.
        pass

    # If we're hot-loading config, do it here!
    if interactively_set_key:
        _interactively_get_key(env)
        return

    # If we're still here, we're not interactively setting the key. Display an error message and exit.
    _report_error_and_exit(env)



import click

import steamship


def requirements_init_wizard():
    click.secho(
        "Steamship uses requirements.txt to specify dependencies. You do not currently have a requirements.txt in this directory.",
        fg="yellow",
    )
    if not click.confirm("Would you like to create one automatically?", default=True):
        click.secho("Please manually create a requirements.txt and try again.")
        click.get_current_context().abort()

    with open("requirements.txt", "w") as requirements_file:
        requirements_file.write(f"steamship=={steamship.__version__}\n")

    click.secho(
        "Created a requirements.txt with the steamship dependency. If you need others, they must be added manually."
    )



import logging

from steamship import SteamshipError, Task
from steamship.agents.logging import AgentLogging
from steamship.agents.schema import Action, Agent, FinishAction
from steamship.agents.schema.context import AgentContext
from steamship.invocable import PackageService


class AgentService(PackageService):
    """AgentService is a Steamship Package that can use an Agent, Tools, and a provided AgentContext to
    respond to user input."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    ###############################################
    # Tool selection / execution
    ###############################################

    def run_action(self, action: Action, context: AgentContext):
        if isinstance(action, FinishAction):
            return

        blocks_or_task = action.tool.run(tool_input=action.input, context=context)
        if isinstance(blocks_or_task, Task):
            raise SteamshipError(
                "Tools return Tasks are not yet supported (but will be soon). "
                "Please use synchronous Tasks (Tools that return List[Block] for now."
            )
        else:
            outputs = ",".join([f"{b.as_llm_input()}" for b in blocks_or_task])
            logging.info(
                f"Tool {action.tool.name}: ({outputs})",
                extra={
                    AgentLogging.TOOL_NAME: action.tool.name,
                    AgentLogging.IS_MESSAGE: True,
                    AgentLogging.MESSAGE_TYPE: AgentLogging.OBSERVATION,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )
            action.output = blocks_or_task
            context.completed_steps.append(action)

    def run_agent(self, agent: Agent, context: AgentContext):
        action = agent.next_action(context=context)
        while not isinstance(action, FinishAction):
            # TODO: Arrive at a solid design for the details of this structured log object
            inputs = ",".join([f"{b.as_llm_input()}" for b in action.input])
            logging.info(
                f"Running Tool {action.tool.name} ({inputs})",
                extra={
                    AgentLogging.TOOL_NAME: action.tool.name,
                    AgentLogging.IS_MESSAGE: True,
                    AgentLogging.MESSAGE_TYPE: AgentLogging.ACTION,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )
            self.run_action(action=action, context=context)
            action = agent.next_action(context=context)
            # TODO: Arrive at a solid design for the details of this structured log object
            logging.info(
                f"Next Tool: {action.tool.name}",
                extra={
                    AgentLogging.TOOL_NAME: action.tool.name,
                    AgentLogging.IS_MESSAGE: False,
                    AgentLogging.MESSAGE_TYPE: AgentLogging.ACTION,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )

        context.completed_steps.append(action)
        for func in context.emit_funcs:
            func(action.output, context.metadata)



from typing import List, Optional

from steamship.agents.tools.text_generation.text_rewrite_tool import TextRewritingTool
from steamship.utils.repl import ToolREPL

DEFAULT_LABELS = ["question", "complaint", "suggestion"]

DEFAULT_PROMPT = """Instructions:
Please classify the following message into one of the following output labels. Respond with exactly one and only one label.

Labels:
{labels}

Passage:
{{input}}

Label describing passage:"""


class ZeroShotClassifierTool(TextRewritingTool):
    """
    Example tool to illustrate how one might classify a user message.

    For example: the agent may wish to know if the use message was a question, complaint, or suggestion.

    TODO: This feels like it wants to emit data to a side channel. Or perhaps it TAGS the user input block?
    """

    name = "ZeroShotClassifierTool"
    human_description = "Classifies a user message."
    agent_description = "Used to classify a user message. The input is a string, and the output is a string with the classification label."
    labels: List[str] = DEFAULT_LABELS
    rewrite_prompt: str = DEFAULT_PROMPT

    def __init__(
        self, labels: Optional[List[str]] = None, rewrite_prompt: Optional[str] = None, **kwargs
    ):
        _rewrite_prompt = rewrite_prompt or DEFAULT_PROMPT
        kwargs["rewrite_prompt"] = kwargs.get(
            "rewrite_prompt", _rewrite_prompt.format(labels=labels or DEFAULT_LABELS)
        )
        super().__init__(**kwargs)


if __name__ == "__main__":
    ToolREPL(ZeroShotClassifierTool()).run()



from __future__ import annotations

from typing import Dict, List, Optional, Union

from steamship import File, MimeTypes, SteamshipError
from steamship.base.client import Client
from steamship.data import TagKind
from steamship.data.block import Block
from steamship.data.tags import Tag
from steamship.data.tags.tag_constants import ChatTag, DocTag, RoleTag, TagValueKey


class ChatHistory:
    """A ChatHistory is a wrapper of a File ideal for ongoing interactions between a user and a virtual assistant."""

    file: File

    def __init__(self, file: File):
        """This init method is intended only for private use within the class. See `Chat.create()`"""
        self.file = file

    @staticmethod
    def _get_existing_file(client: Client, context_keys: Dict[str, str]) -> Optional[File]:
        """Find an existing File object whose memory Tag matches the passed memory keys"""
        file_query = " and ".join(
            [f'value("{key}") = "{value}"' for key, value in context_keys.items()]
        )
        file = File.query(client, file_query)
        if len(file.files) == 1:
            return file.files[0]
        elif len(file.files) == 0:
            return None
        else:
            raise SteamshipError(
                "Multiple ChatHistory objects have been created in this workspace with these memory keys."
            )

    @staticmethod
    def get_or_create(
        client: Client,
        context_keys: Dict[str, str],
        tags: List[Tag] = None,
    ) -> ChatHistory:

        file = ChatHistory._get_existing_file(client, context_keys)

        if file is None:
            tags = tags or []
            tags.append(Tag(kind=TagKind.DOCUMENT, name=DocTag.CHAT))
            tags.append(Tag(kind=TagKind.CHAT, name=ChatTag.CONTEXT_KEYS, value=context_keys))

            blocks = []
            file = File.create(
                client=client,
                blocks=blocks,
                tags=tags,
            )
        return ChatHistory(file)

    def append_user_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with content provided by the end-user."""
        tags = tags or []
        tags.append(
            Tag(
                kind=TagKind.CHAT, name=ChatTag.ROLE, value={TagValueKey.STRING_VALUE: RoleTag.USER}
            )
        )
        return self.file.append_block(
            text=text, tags=tags, content=content, url=url, mime_type=mime_type
        )

    def append_system_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with content provided by the system, i.e., instructions to the assistant."""
        tags = tags or []
        tags.append(
            Tag(
                kind=TagKind.CHAT,
                name=ChatTag.ROLE,
                value={TagValueKey.STRING_VALUE: RoleTag.SYSTEM},
            )
        )
        return self.file.append_block(
            text=text, tags=tags, content=content, url=url, mime_type=mime_type
        )

    def append_agent_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with content provided by the agent, i.e., results from the assistant."""
        tags = tags or []
        tags.append(
            Tag(
                kind=TagKind.CHAT,
                name=ChatTag.ROLE,
                value={TagValueKey.STRING_VALUE: RoleTag.ASSISTANT},
            )
        )
        return self.file.append_block(
            text=text, tags=tags, content=content, url=url, mime_type=mime_type
        )

    @property
    def last_user_message(self) -> Optional[Block]:
        for block in self.file.blocks[::-1]:
            if block.chat_role == RoleTag.USER:
                return block
        return None

    @property
    def last_system_message(self) -> Optional[Block]:
        for block in self.file.blocks[::-1]:
            if block.chat_role == RoleTag.SYSTEM:
                return block
        return None

    @property
    def last_agent_message(self) -> Optional[Block]:
        for block in self.file.blocks[::-1]:
            if block.chat_role == RoleTag.ASSISTANT:
                return block
        return None

    @property
    def initial_system_prompt(self) -> Optional[Block]:
        if len(self.file.blocks) > 0 and self.file.blocks[0].chat_role == RoleTag.SYSTEM:
            return self.file.blocks[0]
        else:
            return None

    def refresh(self):
        self.file.refresh()

    @property
    def tags(self) -> List[Tag]:
        return self.file.tags

    @property
    def messages(self) -> List[Block]:
        return self.file.blocks

    @property
    def client(self) -> Client:
        return self.file.client



from typing import List, Optional

from steamship import Tag


def get_tag_value_key(
    tags: Optional[List[Tag]], key: str, kind: Optional[str] = None, name: Optional[str] = None
) -> Optional[any]:
    """Iterates through a list of tags and returns the first that contains the provided Kind/Name/ValueKey."""
    for tag in tags or []:
        if (kind is None or tag.kind == kind) and (name is None or tag.name == name):
            return (tag.value or {}).get(key)
    return None



from __future__ import annotations

import base64
from typing import Any

from steamship.base.mime_types import TEXT_MIME_TYPES, MimeTypes
from steamship.base.model import CamelModel
from steamship.utils.signed_urls import url_to_bytes


def is_base64(sb):
    # noinspection PyBroadException
    try:
        if isinstance(sb, str):
            # If there's Any unicode here, an exception will be thrown and the function will return false
            sb_bytes = bytes(sb, "ascii")
        elif isinstance(sb, bytes):
            sb_bytes = sb
        else:
            raise ValueError("Argument must be string or bytes")
        return base64.b64encode(base64.b64decode(sb_bytes)) == sb_bytes
    except Exception:
        return False


class RawDataPluginInput(CamelModel):
    """Input for a plugin that accepts raw data, plus a mime type.

    A plugin author need only ever concern themselves with two fields:
    - `data` - Raw bytes
    ` `default_mime_type` - The best guess as to `data`'s MIME Type unless otherwise known to be different.

    In practice, however, the lifecycle of this object involves a bit more under the hood:

    - **Potentially Base64 Decoding Data**. When decoding from a dict, the `data` field is assumed to be Base64 encoded.
      This is to support JSON as a transport encoding over the wire. The constructor automatically performs the
      decoding, and the Steamship Engine automatically performs the encoding, so the Plugin Author can mostly ignore
      this fact.

    - **Potentially late-fetching the `data` from a `url`**. Some files are too large to comfortably send as Base64
      within JSON. The Steamship Engine sometimes chooses to send an empty `data` field paired with a non-empty
      `url` field. When this happens, the constructor proactively, synchronously fetches the contents of that `url`
      and assigns it to the `data` field, throwing a SteamshipError if the fetch fails. Again, this is done
      automatically so the Plugin Author can mostly ignore this fact.
    """

    plugin_instance: str = None
    data: Any = None
    default_mime_type: MimeTypes = None

    def __init__(self, **kwargs):
        data = kwargs.get("data")
        url = kwargs.get("url")

        if data is not None and is_base64(data):
            data_bytes = base64.b64decode(data)
            if kwargs.get("defaultMimeType") in TEXT_MIME_TYPES:
                kwargs["data"] = data_bytes.decode("utf-8")
            else:
                kwargs["data"] = data_bytes
        elif url is not None:
            kwargs["data"] = url_to_bytes(url)  # Resolve the URL into the data field
            kwargs.pop(
                "url"
            )  # Remove the URL field to preserve a simple interface for the consumer

        super().__init__(**kwargs)



from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional

from steamship import SteamshipError, Task
from steamship.invocable import Invocable

# Note!
# =====
#
# This the files in this package are for Package Implementors.
# If you are using the Steamship Client, you probably are looking for either steamship.client or steamship.data
#
from steamship.utils.url import Verb


class PackageService(Invocable):
    """The Abstract Base Class of a Steamship Package.

    Packages may implement whatever methods they like.  To expose these methods as invocable HTTP routes,
    annotate the method with @get or @post and the route name.

    Package *implementations* are effectively stateless, though they will have stateful

    """

    def invoke_later(
        self,
        method: str,
        verb: Verb = Verb.POST,
        wait_on_tasks: List[Task] = None,
        arguments: Dict[str, Any] = None,
        delay_ms: Optional[int] = None,
    ) -> Task[Any]:
        """Schedule a method for future invocation.

        Parameters
        ----------
        method: str
                The method to invoke, as registered with Steamship in the @get or @post decorator.
        verb:   Verb
                The HTTP Verb to use. Default is POST.
        wait_on_tasks: List[Task]
                A list of Task objects (or task IDs) that should be waited upon before invocation.
        arguments: Dict[str, Any]
                The keyword arguments of the invoked method
        delay_ms: Optional[int]
                A delay, in milliseconds, before the invocation should execute.

        Returns
        -------
        Task[Any]
                A Task representing the future work
        """

        if self.context is None:
            raise SteamshipError(
                message="Unable to call invoke_later because the InvocationContext was None"
            )
        if self.context.invocable_instance_handle is None:
            raise SteamshipError(
                message="Unable to call invoke_later because the invocable_instance_handle on InvocationContext was None"
            )

        payload = {
            "instanceHandle": self.context.invocable_instance_handle,
            "payload": {
                "httpVerb": verb.value,
                "invocationPath": method,
                "arguments": arguments or {},
            },
        }
        operation = "package/instance/invoke"

        logging.info(
            f"Scheduling {verb} {method} for future invocation on me ({self.context.invocable_handle})"
        )

        resp = self.client.post(
            operation,
            payload,
            expect=Task[Task],  # This operation should return a task
            as_background_task=True,  # This operation should always be asynchronous
            wait_on_tasks=wait_on_tasks,  # This operation might await other tasks first
            task_delay_ms=delay_ms,  # This operation might have a required delay
        )
        return resp



import pytest

from steamship import Steamship
from steamship.experimental.easy import scrape


@pytest.mark.usefixtures("client")
def test_import_wikipedia(client: Steamship):
    file = scrape(client, "https://en.wikipedia.org/wiki/Steamship")
    assert file.id is not None
    content = file.raw().decode("utf-8")
    assert "Steamship" in content
    file.delete()


# @pytest.mark.usefixtures("client")
# def test_import_youtube(client: Steamship):
#     file = scrape(client, "https://www.youtube.com/watch?v=LXDZ6aBjv_I")
#     assert file.id is not None
#     content = file.raw()
#     assert content is not None
#     file.delete()



import pytest
from steamship_tests.utils.fixtures import get_steamship_client
from steamship_tests.utils.random import random_index, random_name

from steamship import SteamshipError, Tag
from steamship.base.request import SortOrder
from steamship.base.tasks import TaskComment
from steamship.data.embeddings import EmbeddedItem

_TEST_EMBEDDER = "test-embedder"


def _list_equal(actual, expected):
    assert len(actual) == len(expected)
    assert all([a == b for a, b in zip(actual, expected)])


def test_basic_task_comment():
    steamship = get_steamship_client()
    with random_index(steamship, _TEST_EMBEDDER) as index:
        item1 = Tag(text="Pizza", name="pizza", kind="food", value={"value": [1, 2, 3]})

        index.insert(item1)

        res2 = index.search(item1.text, k=1)
        res2.add_comment(external_id="Foo", external_type="Bar", metadata={"value": [1, 2]})
        # We don't return to Res2 until the end to make sure we aren't co-mingling comments!

        res = index.search(item1.text, k=1)
        res.wait()
        items = res.output.items
        assert items is not None
        assert len(items) == 1
        assert items[0].tag.text == item1.text
        assert items[0].tag.name == item1.name
        assert items[0].tag.kind == item1.kind
        _list_equal(items[0].tag.value.get("value"), item1.value.get("value"))

        res.add_comment(external_id="Foo", external_type="Bar", metadata=[1, 2])

        comments = TaskComment.list(client=steamship, task_id=res.task_id)
        assert len(comments.comments) == 1

        comment = comments.comments[0]
        assert comment.external_id == "Foo"
        assert comment.external_type == "Bar"
        _list_equal(comment.metadata, [1, 2])

        comment.delete()

        comments = TaskComment.list(client=steamship, task_id=res.task_id)
        assert len(comments.comments) == 0

        # Now let's add one
        res.add_comment(external_id="Foo1", external_type="Bar1", metadata=[1, 2, 3])
        res.add_comment(external_id="Foo2", external_type="Bar2", metadata=[1, 2, 3, 4])

        comments = TaskComment.list(client=steamship, task_id=res.task_id, sort_order=SortOrder.ASC)
        assert len(comments.comments) == 2
        comment = comments.comments[0]
        assert comment.external_id == "Foo1"
        assert comment.external_type == "Bar1"
        _list_equal(comment.metadata, [1, 2, 3])

        comment = comments.comments[1]
        assert comment.external_id == "Foo2"
        assert comment.external_type == "Bar2"
        _list_equal(comment.metadata, [1, 2, 3, 4])

        comments.comments[0].delete()
        comments.comments[1].delete()

        comments = TaskComment.list(client=steamship, task_id=res.task_id)
        assert len(comments.comments) == 0

        # Now we handle res2
        comments = TaskComment.list(client=steamship, task_id=res2.task_id)
        assert len(comments.comments) == 1
        comment = comments.comments[0]
        assert comment.external_id == "Foo"
        assert comment.external_type == "Bar"
        _list_equal(comment.metadata, {"value": [1, 2]})
        comments.comments[0].delete()
        comments = TaskComment.list(client=steamship, task_id=res.task_id)
        assert len(comments.comments) == 0


def test_task_comment_list_paging():
    steamship = get_steamship_client()
    with random_index(steamship, _TEST_EMBEDDER) as index:
        item1 = Tag(text="Pizza", name="pizza", kind="food", value={"value": [1, 2, 3]})
        index.insert(item1)

        res = index.search(item1.text, k=1)
        res.wait()

        res.add_comment(external_id="Foo1", external_type="Bar1", metadata=[1, 2, 3])
        res.add_comment(external_id="Foo2", external_type="Bar2", metadata=[1, 2, 3, 4])
        res.add_comment(external_id="Foo3", external_type="Bar3", metadata=[1, 2, 3, 4])

        resp = TaskComment.list(client=steamship, task_id=res.task_id, page_size=2)
        assert len(resp.comments) == 2
        assert resp.next_page_token is not None

        resp = TaskComment.list(
            client=steamship, task_id=res.task_id, page_size=2, page_token=resp.next_page_token
        )
        assert len(resp.comments) == 1
        assert resp.next_page_token is None

        with pytest.raises(SteamshipError):
            TaskComment.list(steamship, page_size=-1)


def test_task_comment_feedback_reporting():
    """
    We want to be able to generate reports like this:

    Select Across Gorup    -- externalGroup
    Inputs Seen: XXX       -- Distinct externalId
    Inputs Suggested: YYY  -- Add to metadata
    Inputs Liked / Disliked / Used -- Add to metadata

    So really we just need to test the group aggregation
    """
    client = get_steamship_client()
    with random_index(client, plugin_instance=_TEST_EMBEDDER) as index:
        item1 = EmbeddedItem(
            value="Pizza", external_id="pizza", external_type="food", metadata={"value": [1, 2, 3]}
        )

        group_name_1 = random_name()
        group_name_2 = random_name()

        index.insert(
            Tag(
                text=item1.value,
                name=item1.external_id,
                kind=item1.external_type,
                value=item1.metadata,
            )
        )

        res = index.search(item1.value, k=1)
        res.add_comment(
            external_id="Foo1",
            external_type="Bar1",
            external_group=group_name_1,
            metadata=[1, 2, 3],
        )
        res.add_comment(
            external_id="Foo2",
            external_type="Bar1",
            external_group=group_name_1,
            metadata=[1, 2, 3],
        )
        res.add_comment(
            external_id="Foo2",
            external_type="Bar1",
            external_group=group_name_2,
            metadata=[1, 2, 3],
        )

        comments = TaskComment.list(client=client, task_id=res.task_id)
        assert len(comments.comments) == 3

        g1 = TaskComment.list(
            client=client,
            external_group=group_name_1,
        )
        assert len(g1.comments) == 2

        g2 = TaskComment.list(client=client, external_group=group_name_2)
        assert len(g2.comments) == 1

        g1 = TaskComment.list(client=client, task_id=res.task_id, external_group=group_name_1)
        assert len(g1.comments) == 2

        g2 = TaskComment.list(client=client, task_id=res.task_id, external_group=group_name_2)
        assert len(g2.comments) == 1

        g1 = TaskComment.list(
            client=client, task_id=res.task_id, external_id="Foo1", external_group=group_name_1
        )
        assert len(g1.comments) == 1

        g2 = TaskComment.list(
            client=client, task_id=res.task_id, external_id="Foo1", external_group=group_name_2
        )
        assert len(g2.comments) == 0

        comments.comments[0].delete()
        comments.comments[1].delete()
        comments.comments[2].delete()

        g1 = TaskComment.list(client=client, external_group=group_name_1)
        assert len(g1.comments) == 0

        g2 = TaskComment.list(client=client, external_group=group_name_2)
        assert len(g2.comments) == 0



import uuid
from typing import List

from steamship import Block
from steamship.agents.llms.openai import OpenAI
from steamship.agents.react import ReACTAgent
from steamship.agents.schema import AgentContext
from steamship.agents.schema.context import Metadata
from steamship.agents.service.agent_service import AgentService
from steamship.agents.tools.image_generation import DalleTool
from steamship.agents.tools.search import SearchTool
from steamship.invocable import post
from steamship.utils.repl import AgentREPL


class MyAssistant(AgentService):
    """MyAssistant is an example AgentService that exposes a single test endpoint
    for trying out Agent-based invocations. It is configured with two simple Tools
    to provide an overview of the types of tasks it can accomplish (here, search
    and image generation)."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._agent = ReACTAgent(
            tools=[
                SearchTool(),
                DalleTool(),
            ],
            llm=OpenAI(self.client),
        )

    @post("prompt")
    def prompt(self, prompt: str) -> str:
        """Run an agent with the provided text as the input."""

        # AgentContexts serve to allow the AgentService to run agents
        # with appropriate information about the desired tasking.
        # Here, we create a new context on each prompt, and append the
        # prompt to the message history stored in the context.
        context_id = uuid.uuid4()
        context = AgentContext.get_or_create(self.client, {"id": f"{context_id}"})
        context.chat_history.append_user_message(prompt)

        # AgentServices provide an emit function hook to access the output of running
        # agents and tools. The emit functions fire at after the supplied agent emits
        # a "FinishAction".
        #
        # Here, we show one way of accessing the output in a synchronous fashion. An
        # alternative way would be to access the final Action in the `context.completed_steps`
        # after the call to `run_agent()`.
        output = ""

        def sync_emit(blocks: List[Block], meta: Metadata):
            nonlocal output
            block_text = "\n".join(
                [b.text if b.is_text() else f"({b.mime_type}: {b.id})" for b in blocks]
            )
            output += block_text

        context.emit_funcs.append(sync_emit)
        self.run_agent(self._agent, context)
        return output


if __name__ == "__main__":
    # AgentREPL provides a mechanism for local execution of an AgentService method.
    # This is used for simplified debugging as agents and tools are developed and
    # added.
    AgentREPL(MyAssistant, "prompt", agent_package_config={}).run()



import pytest
import requests
from assets.plugins.generators.test_generator_returns_bytes import TEST_BYTES_STRING
from steamship_tests import PLUGINS_PATH, TEST_ASSETS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import Block, File, MimeTypes, PluginInstance, Steamship


def test_e2e_generator():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "generators" / "test_generator.py"

    with deploy_plugin(client, parser_path, "generator") as (
        plugin,
        version,
        instance,
    ):
        test_doc = "Yo! Banana boy!"
        res = instance.generate(text=test_doc, append_output_to_file=False)
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 1
        assert res.output.blocks[0].text == "!yob ananaB !oY"


def test_e2e_generator_with_existing_file():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "generators" / "test_generator.py"

    with deploy_plugin(client, parser_path, "generator") as (
        plugin,
        version,
        instance,
    ):

        # do an all-blocks test
        test_file = File.create(client, blocks=[Block(text="Yo! Banana boy!")])
        res = instance.generate(
            input_file_id=test_file.id, append_output_to_file=True, output_file_id=test_file.id
        )
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 1
        assert res.output.blocks[0].text == "!yob ananaB !oY"

        test_file.refresh()
        assert len(test_file.blocks) == 2
        assert test_file.blocks[1].text == "!yob ananaB !oY"

        # do a some-blocks test
        res = instance.generate(
            input_file_id=test_file.id,
            input_file_start_block_index=1,
            input_file_end_block_index=2,
            append_output_to_file=True,
            output_file_id=test_file.id,
        )
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 1
        assert res.output.blocks[0].text == "Yo! Banana boy!"

        test_file.refresh()
        assert len(test_file.blocks) == 3
        assert test_file.blocks[2].text == "Yo! Banana boy!"

        # do an index-list test
        test_file = File.create(
            client,
            blocks=[
                Block(text="Yo! Banana boy!"),
                Block(text="A man, a plan, a canal, Panama!"),
                Block(text="Tacocat!"),
            ],
        )
        res = instance.generate(input_file_id=test_file.id, input_file_block_index_list=[0, 2])
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 2
        assert res.output.blocks[0].text == "!yob ananaB !oY"
        assert res.output.blocks[1].text == "!tacocaT"


def test_e2e_generator_runtime_options():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "generators" / "test_generator.py"

    with deploy_plugin(client, parser_path, "generator") as (
        plugin,
        version,
        instance,
    ):
        # do an all-blocks test
        test_file = File.create(client, blocks=[Block(text="Yo! Banana boy!")])
        res = instance.generate(
            input_file_id=test_file.id,
            append_output_to_file=True,
            output_file_id=test_file.id,
            options={"test": "yes"},
        )
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 2
        assert res.output.blocks[0].text == "!yob ananaB !oY"
        assert res.output.blocks[1].text == '{"test": "yes"}'


def test_generator_ephemeral_image_output():
    client = get_steamship_client()
    plugin_instance = PluginInstance.create(client, plugin_handle="test-image-generator")
    generate_task = plugin_instance.generate(text="This won't be used")

    generate_task.wait()
    assert generate_task.output is not None
    assert len(generate_task.output.blocks) == 1
    assert generate_task.output.blocks[0].content_url is not None
    data = generate_task.output.blocks[0].raw()
    assert data.decode("UTF-8") == "PRETEND THIS IS THE DATA OF AN IMAGE"


def test_e2e_generate_from_image():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "generators" / "test_image_to_text_generator.py"

    with deploy_plugin(client, parser_path, "generator") as (
        _,
        _,
        generator,
    ):
        test_file = File.create(client, content="")
        palm_tree_path = TEST_ASSETS_PATH / "palm_tree.png"

        with palm_tree_path.open("rb") as f:
            palm_bytes = f.read()
        block = Block.create(
            client, file_id=test_file.id, content=palm_bytes, mime_type=MimeTypes.PNG
        )

        fetched_bytes = block.raw()

        assert palm_bytes == fetched_bytes

        generate_task = generator.generate(input_file_id=test_file.id)
        generate_task.wait()
        result = generate_task.output.blocks[0].text
        assert result == "Found 1 image blocks and fetched data from 1"


def test_e2e_generate_returning_bytes():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "generators" / "test_generator_returns_bytes.py"

    with deploy_plugin(client, parser_path, "generator") as (
        _,
        _,
        generator,
    ):

        # Test ephemerally
        generate_task = generator.generate(text="")
        generate_task.wait()

        ephemeral_block = generate_task.output.blocks[0]
        assert ephemeral_block.content_url is not None
        ephemeral_block_content = ephemeral_block.raw()
        assert ephemeral_block_content == TEST_BYTES_STRING.encode("utf-8")

        # test persisted
        generate_task_persist = generator.generate(text="", append_output_to_file=True)
        generate_task_persist.wait()
        persisted_block = generate_task_persist.output.blocks[0]
        assert persisted_block.content_url is None
        assert persisted_block.file_id is not None
        persisted_block_content = persisted_block.raw()
        assert persisted_block_content == TEST_BYTES_STRING.encode("utf-8")


@pytest.mark.usefixtures("client")
def test_generate_block_public_data(client: Steamship):
    client = get_steamship_client()
    plugin_instance = PluginInstance.create(client, plugin_handle="test-image-generator")
    generate_task = plugin_instance.generate(
        text="This won't be used", append_output_to_file=True, make_output_public=True
    )

    blocks = generate_task.wait().blocks
    assert blocks is not None
    assert blocks[0].public_data

    # Intentionally no API key
    response = requests.get(blocks[0].raw_data_url)

    assert response.text == "PRETEND THIS IS THE DATA OF AN IMAGE"
    assert response.headers["content-type"] == MimeTypes.PNG


@pytest.mark.usefixtures("client")
def test_generate_block_private_data(client: Steamship):
    client = get_steamship_client()
    plugin_instance = PluginInstance.create(client, plugin_handle="test-image-generator")
    generate_task = plugin_instance.generate(text="This won't be used", append_output_to_file=True)

    blocks = generate_task.wait().blocks
    assert blocks is not None
    assert not blocks[0].public_data

    # Intentionally no API key
    failed_response = requests.get(blocks[0].raw_data_url)
    assert not failed_response.ok

    # Should still be able to get with API key
    response = requests.get(
        blocks[0].raw_data_url,
        headers={"Authorization": f"Bearer {client.config.api_key.get_secret_value()}"},
    )

    assert response.text == "PRETEND THIS IS THE DATA OF AN IMAGE"
    assert response.headers["content-type"] == MimeTypes.PNG



from enum import Enum
from typing import Optional


class TagKind(str, Enum):
    """A set of `kind` constants for Tags.

    These define broad categories of tags. Suggested `name` values for each category are found in
    separate enums. For example: kind=TagKind.DOCUMENT, name=DocTag.H1
    """

    PART_OF_SPEECH = "part-of-speech"
    DEPENDENCY = "dependency"
    SENTIMENT = "sentiment"
    EMOTION = "emotion"
    ENTITY = "entity"
    DOCUMENT = "document"
    TOKEN = "token"  # noqa: S105
    INTENT = "intent"
    EMBEDDING = "embedding"
    GENERATION = "generation"
    PROVENANCE = "provenance"
    TOPIC = "topic"
    TOKENIZATION = "tokenization"
    KIND = "summary"
    TIMESTAMP = "timestamp"
    SUMMARY = "summary"
    SEARCH_RESULT = "search-result"
    ROLE = "role"
    CHAT = "chat"
    CHAT_HISTORY_CONTEXT = "chat-history-context"
    MESSAGE_ID = "message-id"


class DocTag(str, Enum):
    """A set of `name` constants for Tags with a `kind` of `TagKind.doc`; appropriate for HTML and Markdown ideas."""

    DOCUMENT = "document"
    PAGE = "page"  # E.g. in a PDF
    REGION = "region"  # E.g., abstract catchall region in a document
    HEADER = "header"
    H1 = "h1"
    H2 = "h2"
    H3 = "h3"
    H4 = "h4"
    H5 = "h5"
    LINE = "line"
    TITLE = "title"
    SUBTITLE = "subtitle"
    FOOTER = "footer"
    PARAGRAPH = "paragraph"
    ORDERED_LIST = "ordered-list"
    UNORDERED_LIST = "unordered-list"
    LIST_ITEM = "list-item"
    LINK = "link"
    CAPTION = "caption"
    IMAGE = "image"
    BLOCK_QUOTE = "block-quote"
    BLOCK_CODE = "block-code"
    UNKNOWN = "unknown"
    SENTENCE = "sentence"
    TOKEN = "token"  # noqa: S105
    SPAN = "span"
    DIV = "div"
    PRE = "pre"
    STRONG = "strong"
    EMPHASIZED = "emphasized"
    UNDERLINED = "underlined"
    TELETYPE = "teletype"
    ARTICLE = "article"
    MAIN = "main"
    CHAPTER = "chapter"
    TEXT = "text"
    CHAT = "chat"

    @staticmethod
    def from_html_tag(tagname: Optional[str]) -> Optional["DocTag"]:  # noqa: C901
        if tagname is None:
            return None

        name = tagname.lower().strip()

        if name == "p":
            return DocTag.PARAGRAPH
        elif name == "h1":
            return DocTag.H1
        elif name == "h2":
            return DocTag.H2
        elif name == "h3":
            return DocTag.H3
        elif name == "h4":
            return DocTag.H4
        elif name == "h5":
            return DocTag.H5
        elif name == "ul":
            return DocTag.UNORDERED_LIST
        elif name == "ol":
            return DocTag.ORDERED_LIST
        elif name == "li":
            return DocTag.LIST_ITEM
        elif name == "a":
            return DocTag.LINK
        elif name == "div":
            return DocTag.DIV
        elif name == "img":
            return DocTag.IMAGE
        elif name == "span":
            return DocTag.SPAN
        elif name == "pre":
            return DocTag.PRE
        elif name == "code":
            return DocTag.BLOCK_CODE
        elif name == "blockquote":
            return DocTag.BLOCK_QUOTE
        elif name == "strong":
            return DocTag.STRONG
        elif name == "b":
            return DocTag.STRONG
        elif name == "emph":
            return DocTag.EMPHASIZED
        elif name == "i":
            return DocTag.EMPHASIZED
        elif name == "u":
            return DocTag.UNDERLINED
        elif name == "tt":
            return DocTag.TELETYPE
        elif name == "article":
            return DocTag.ARTICLE
        elif name == "header":
            return DocTag.HEADER
        elif name == "footer":
            return DocTag.FOOTER
        elif name == "main":
            return DocTag.MAIN

        return None


class TokenTag(str, Enum):
    """A set of `name` constants for Tags with a `kind` of `TagKind.token`; appropriate for parsing-level ideas."""

    TEXT_WITH_WHITESPACE = "text-with-whitespace"
    TEXT = "text"
    WHITESPACE = "whitespace"
    HEAD = "head"
    LEFT_EDGE = "left-edge"
    RIGHT_EDGE = "right-edge"
    ENTITY_TYPE = "entity-type"
    ENTITY_IOB = "entity-iob"
    LEMMA = "lemma"
    NORMALIZED = "normalized"
    SHAPE = "shape"
    PREFIX = "prefix"
    SUFFIX = "suffix"
    IS_ALPHA = "is-alpha"
    IS_ASCII = "is-ascii"
    IS_DIGIT = "is-digit"
    IS_TITLE = "is-title"
    IS_PUNCT = "is-punct"
    IS_LEFT_PUNCT = "is-left-punct"
    IS_RIGHT_PUNCT = "is-right-punct"
    IS_SPACE = "is-space"
    IS_BRACKET = "is-bracket"
    IS_QUOTE = "is-quote"
    IS_CURRENCY = "is-currency"
    LIKE_URL = "like-url"
    LIKE_NUM = "like-num"
    LIKE_EMAIL = "like-email"
    IS_OUT_OF_VOCABULARY = "is-out-of-vocabulary"
    IS_STOPWORD = "is-stopword"
    LANGUAGE = "language"


class TagValueKey(str, Enum):
    """A set of key constants for the `value` object within a tag."""

    # Catch-all for confidence, score, ranking
    SCORE = "score"

    # Catch-all for values of different types such as integers, floats, booleans, and strings
    VALUE = "value"

    # An array of floats or integers
    VECTOR_VALUE = "vector-value"

    # A float or integer
    NUMBER_VALUE = "number-value"

    # A bool
    BOOL_VALUE = "bool-value"

    # A string
    STRING_VALUE = "string-value"

    # Whether some annotation is direct ("Susan said 'Hi'")
    DIRECT = "direct"

    # Start time of a region of a document, in some other medium (seconds)
    START_TIME_S = "start-time-s"

    # End time of a region of a document, in some other medium (seconds)
    END_TIME_S = "end-time-s"

    # The normalized name of an entity
    ENTITY_NAME = "entity_name"

    # Timestamp. Can be used to provide a time-based sort-ordering for tags.
    TIMESTAMP_VALUE = "timestamp-value"


class GenerationTag(str, Enum):
    """A set of `name` constants for Tags with a `kind` of `TagKind.generation`."""

    # A generated summary of some region of a document
    SUMMARY = "summary"

    # A generated headline for some region of a document
    HEADLINE = "headline"

    # A generated "micro summary" of some region of a document
    GIST = "gist"

    # A generated completion using some region of the document as input
    PROMPT_COMPLETION = "prompt-completion"


class ProvenanceTag(str, Enum):
    """A set of `name` constants for Tags with a `kind` of `TagKind.provenance`."""

    # The speaker of a section of a document
    SPEAKER = "speaker"

    # The URL from which some section of a document was sourced
    URL = "url"

    # The File from which some section of a document was sourced
    FILE = "file"


class RoleTag(str, Enum):
    """A set of `name` constants for Tags with a `kind` of `TagKind.ROLE`."""

    # This block's content was created by the System; likely instructional text on how to respond
    SYSTEM = "system"

    # This block's content was created by an end user
    USER = "user"

    # This block's content was created by the generative AI assistant
    ASSISTANT = "assistant"

    # This block's content was created by a non-human agent participating in the chat
    AGENT = "agent"


class ChatTag(str, Enum):
    """A set of `name` constants for Tags with a `kind` of `TagKind.CHAT`."""

    # The chat id in which a message happened
    CHAT_ID = "chat-id"

    # The message id of a message
    MESSAGE_ID = "message-id"

    # The role of a message
    ROLE = "role"

    # The keys to look up a context
    CONTEXT_KEYS = "context-keys"



from steamship import SteamshipError
from steamship.invocable import Invocable, InvocableResponse, post


class PackageWithFailingInstanceInit(Invocable):
    def instance_init(self):
        raise SteamshipError("oh noes I failed")

    @post("say_ok")
    def say_ok(self) -> InvocableResponse:
        return InvocableResponse(data="OK")



from steamship import Steamship
from steamship.agents.llms import OpenAI
from steamship.agents.tools.text_generation.text_rewrite_tool import TextRewritingTool
from steamship.agents.utils import with_llm
from steamship.utils.repl import ToolREPL

DEFAULT_PROMPT = """Instructions:
Please summarize the following document so that the key points are represented in a way that stand alone for reading.

DOCUMENT
========

{input}

ONE PARAGRAPH SUMMARY
====================="""


class SummarizeTextWithPromptTool(TextRewritingTool):
    """
    Example tool to illustrate summarizing an input document using prompt engineering.
    """

    name: str = "SummarizeTextWithPromptTool"
    human_description: str = "Summarizes text using an LLM prompt."
    agent_description: str = (
        "Use this tool to summarize text. "
        "The input is the text needing summarization. "
        "The output is a summary of the text."
    )
    rewrite_prompt: str = DEFAULT_PROMPT


if __name__ == "__main__":
    with Steamship.temporary_workspace() as client:
        ToolREPL(SummarizeTextWithPromptTool()).run_with_client(
            client=client, context=with_llm(llm=OpenAI(client=client, temperature=0.9))
        )



"""Blockifies a file using a best effort guess based on mime-type."""
from typing import Optional

from steamship import MimeTypes, SteamshipError, Task
from steamship.data.file import File


def blockify(file: File, mime_type: Optional[MimeTypes] = None) -> Task:
    """Scrape a file via URL, returning a File object no matter what."""

    _mime_type = mime_type or file.mime_type
    if not _mime_type:
        raise SteamshipError(
            message=f"No MIME Type was found for file {file.id}. Unable to easy-blockify"
        )

    if _mime_type == MimeTypes.PDF:
        plugin_instance = file.client.use_plugin("pdf-blockifier")
    elif _mime_type in [MimeTypes.MP3, MimeTypes.MP4_AUDIO, MimeTypes.WEBM_AUDIO]:
        plugin_instance = file.client.use_plugin("s2t-blockifier-default")
    elif _mime_type in [MimeTypes.MKD, MimeTypes.TXT]:
        plugin_instance = file.client.use_plugin("markdown-blockifier-default")
    else:
        raise SteamshipError(
            message=f"Unable to easy.blockfy file of MIME type {mime_type}. File ID was {file.id}"
        )

    return file.blockify(plugin_instance.handle)



from enum import Enum


class HostingType(str, Enum):
    """The type of hosting provider to deploy to."""

    LAMBDA = "lambda"
    ECS = "ecs"


class HostingEnvironment(str, Enum):
    """The software environment required for deployment."""

    PYTHON38 = "python38"
    STEAMSHIP_PYTORCH_CPU = "inferenceCpu"


class HostingMemory(str, Enum):
    """The amount of memory required for deployment.

    This is mapped to a value dependent on the HostingType it is combined with.
    """

    MIN = "min"
    XXS = "xxs"
    XS = "xs"
    SM = "sm"
    MD = "md"
    LG = "lg"
    XL = "xl"
    XXL = "xxl"
    MAX = "max"


class HostingCpu(str, Enum):
    """The amount of CPU required for deployment.

    This is mapped to a value dependent on the HostingType it is combined with.
    """

    MIN = "min"
    XXS = "xxs"
    XS = "xs"
    SM = "sm"
    MD = "md"
    LG = "lg"
    XL = "xl"
    XXL = "xxl"
    MAX = "max"


class HostingTimeout(str, Enum):
    """The request timeout required for deployment.

    This is mapped to a value dependent on the HostingType it is combined with.
    """

    MIN = "min"
    XXS = "xxs"
    XS = "xs"
    SM = "sm"
    MD = "md"
    LG = "lg"
    XL = "xl"
    XXL = "xxl"
    MAX = "max"



from abc import ABC, abstractmethod
from typing import List

from pydantic import BaseModel

from steamship.agents.schema.action import Action
from steamship.agents.schema.context import AgentContext
from steamship.agents.schema.llm import LLM
from steamship.agents.schema.output_parser import OutputParser
from steamship.agents.schema.tool import Tool


class Agent(BaseModel, ABC):
    """Agent is responsible for choosing the next action to take for an AgentService.

    It uses the provided context, and a set of Tools, to decide on an action that will
    be executed by the AgentService.
    """

    tools: List[Tool]
    """Tools that can be used by the Agent in selecting the next Action."""

    @abstractmethod
    def next_action(self, context: AgentContext) -> Action:
        pass


class LLMAgent(Agent):
    """LLMAgents choose next actions for an AgentService based on interactions with an LLM."""

    llm: LLM
    """The LLM to use for action selection."""

    output_parser: OutputParser
    """Utility responsible for converting LLM output into Actions"""

    @abstractmethod
    def next_action(self, context: AgentContext) -> Action:
        pass



import msgpack

from steamship import SteamshipError
from steamship.invocable.lambda_handler import encode_exception


def test_serialize_steamship_error():
    error = SteamshipError("Oh noes! An error happened!")
    serialized = msgpack.packb(error, default=encode_exception)
    assert (
        serialized
        == b'\xd9t{"message": "Oh noes! An error happened!", "internalMessage": null, "suggestion": null, "code": null, "error": null}'
    )


def test_serialize_steamship_error_with_nested_error():
    nested_error = AttributeError("A nested error!")
    error = SteamshipError("Oh noes! An error happened!", error=nested_error)
    serialized = msgpack.packb(error, default=encode_exception)
    assert (
        serialized
        == b'\xd9\x81{"message": "Oh noes! An error happened!", "internalMessage": null, "suggestion": null, "code": null, "error": "A nested error!"}'
    )


def test_serialize_non_steamship_error():
    error = AttributeError("An attribute error!")
    serialized = msgpack.packb(error, default=encode_exception)
    assert serialized == b"\xd9?exception_class: AttributeError, args: ('An attribute error!',)"



import json
from enum import Enum
from pathlib import Path
from typing import Dict

from steamship.base.error import SteamshipError
from steamship.base.model import CamelModel
from steamship.data.manifest import ConfigParameter, ConfigParameterType


class Config(CamelModel):
    """Base class Steamship Package and Plugin configuration objects."""

    def __init__(self, **kwargs):
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        super().__init__(**kwargs)

    def extend_with_dict(self, d: dict, overwrite: bool = False):
        """Sets the attributes on this object with provided keys and values."""
        for key, val in (d or {}).items():
            if hasattr(self, key) and (overwrite or getattr(self, key) is None):
                setattr(self, key, val)

    def extend_with_json_file(
        self, path: Path, overwrite: bool = False, fail_on_missing_file: bool = True
    ):
        """Extends this config object's values with a JSON file from disk.

        This is useful for applying late-bound defaults, such as API keys added to a deployment bundle."""
        if not path.exists():
            if fail_on_missing_file:
                raise SteamshipError(
                    message=f"Attempted to extend Config object with {path}, but the file was not found."
                )
            return

        with open(path) as f:
            data = json.load(f)
            if not isinstance(data, dict):
                raise SteamshipError(
                    message=f"Attempted to extend Config object with {path}, but the file did not contain a JSON `dict` object."
                )
            self.extend_with_dict(data, overwrite)

    @staticmethod
    def strip_enum(default_value):
        if issubclass(type(default_value), Enum):
            return default_value.value
        else:
            return default_value

    @classmethod
    def get_config_parameters(cls) -> Dict[str, ConfigParameter]:
        result = {}
        for field_name, field in cls.__fields__.items():
            description = field.field_info.description
            type_ = ConfigParameterType.from_python_type(field.type_)
            result[field_name] = ConfigParameter(
                type=type_,
                default=cls.strip_enum(field.default),
                description=description,
            )

        return result



import pytest
from assets.packages.demo_package import TestPackage

ERROR_NO_METHOD = "No handler for POST method_doesnt_exist available."
ERROR_STEAMSHIP_ERROR = "[ERROR - POST raise_steamship_error] raise_steamship_error"
ERROR_PYTHON_ERROR = "[ERROR - POST raise_python_error] raise_python_error"


@pytest.mark.parametrize("invocable_handler", [TestPackage], indirect=True)
def test_instance_invoke_unit(invocable_handler):
    """Test that the handler returns the proper errors"""

    response = invocable_handler("POST", "method_doesnt_exist")
    assert response.get("status", {}).get("statusMessage", "") == ERROR_NO_METHOD

    response = invocable_handler("POST", "raise_steamship_error")
    assert response.get("status", {}).get("statusMessage", "") == ERROR_STEAMSHIP_ERROR

    response = invocable_handler("POST", "raise_python_error")
    assert response.get("status", {}).get("statusMessage", "") == ERROR_PYTHON_ERROR



"""Objects for recording and reporting upon the introspected interface of a Steamship Package."""
import inspect
import logging
from copy import deepcopy
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Union, get_args, get_origin

from pydantic import Field

import steamship
from steamship import SteamshipError
from steamship.base.configuration import CamelModel
from steamship.utils.url import Verb


class ArgSpec(CamelModel):
    """An argument passed to a method."""

    # The name of the argument.
    name: str
    # The kind of the argument, reported by str(annotation) via the `inspect` library. E.g. <class 'int'>
    kind: str
    # Possible values, if the kind is an enum type
    values: Optional[List[str]]

    def __init__(self, name: str, parameter: inspect.Parameter):
        if name == "self":
            raise SteamshipError(
                message="Attempt to interpret the `self` object as a method parameter."
            )
        values = None
        if isinstance(parameter.annotation, type):
            if issubclass(parameter.annotation, Enum):
                values = [choice.value for choice in parameter.annotation]
        elif get_origin(parameter.annotation) is Union:
            args = get_args(parameter.annotation)
            # For now, only deal with the case where the Union is an Optional[Enum]
            if len(args) == 2 and type(None) in args:
                optional_arg = [t for t in args if t != type(None)][0]  # noqa: E721
                if issubclass(optional_arg, Enum):
                    values = [choice.value for choice in optional_arg]

        super().__init__(name=name, kind=str(parameter.annotation), values=values)

    def pprint(self, name_width: Optional[int] = None, prefix: str = "") -> str:
        """Returns a pretty printable representation of this argument."""
        width = name_width or len(self.name)
        ret = f"{prefix}{self.name.ljust(width)} - {self.kind}"
        return ret


class MethodSpec(CamelModel):
    """A method, callable remotely, on an object."""

    # The HTTP Path at which the method is callable.
    path: str

    # The HTTP Verb at which the method is callable. Defaults to POST
    verb: str

    # The return type. Reported by str(annotation) via the `inspect` library. E.g. <class 'int'>
    returns: str

    # The docstring of the method.
    doc: Optional[str] = None

    # The named arguments of the method. Positional arguments are not permitted.
    args: Optional[List[ArgSpec]] = None

    # Additional configuration around this endpoint.
    # Note: The actual type of this is Optional[Dict[str, Union[str, bool, int, float]]]
    #       But if Pydantic sees that, it attempts to force all values to be str, which is wrong.
    config: Optional[Dict] = None

    # A bound function to call.
    # If String: the name of a method to call upon a runtime-provided Invocable.
    # If Callable: a function -- on any object -- to call.
    func_binding: Optional[Union[str, Callable[..., Any]]] = Field(None, exclude=True, repr=False)

    @staticmethod
    def clean_path(path: str = "") -> str:
        """Ensure that the path always starts with /, and at minimum must be at least /."""
        if not path:
            path = "/"
        elif path[0] != "/":
            path = f"/{path}"

        if path.startswith("//"):
            path = path[1:]

        return path

    def __init__(self, **kwargs):
        """Create a new instance, making sure the path is properly formatted."""
        if "path" not in kwargs:
            if "name" in kwargs:
                kwargs["path"] = f"{kwargs['name']}"
            else:
                kwargs["path"] = "/"

        # Make sure we sanitize the path to avoid, eg, double //
        kwargs["path"] = MethodSpec.clean_path(kwargs["path"])

        super().__init__(**kwargs)

    @staticmethod
    def from_class(
        cls: object,
        name: str,
        path: str = None,
        verb: Verb = Verb.POST,
        config: Dict[str, Union[str, bool, int, float]] = None,
        func_binding: Optional[Union[str, Callable[..., Any]]] = None,
    ):
        # Get the function on the class so that we can inspect it
        func = getattr(cls, name)
        sig = inspect.signature(func)

        # Set the return type
        returns = str(sig.return_annotation)

        # Set the docstring
        doc = func.__doc__

        # Set the arguments
        args = []
        for p in sig.parameters:
            if p == "self":
                continue
            args.append(ArgSpec(p, sig.parameters[p]))

        return MethodSpec(
            path=path,
            verb=verb,
            returns=returns,
            doc=doc,
            args=args,
            config=config,
            func_binding=func_binding,
        )

    def clone(self) -> "MethodSpec":
        return MethodSpec(
            path=deepcopy(self.path),
            verb=deepcopy(self.verb),
            returns=deepcopy(self.returns),
            doc=deepcopy(self.doc),
            args=deepcopy(self.args),
            config=deepcopy(self.config),
            func_binding=self.func_binding,
        )

    def pprint(self, name_width: Optional[int] = None, prefix: str = "  ") -> str:
        """Returns a pretty printable representation of this method."""

        width = name_width or len(self.path)
        ret = f"{self.verb.ljust(4)} {self.path.lstrip('/').ljust(width)} -> {self.returns}"
        if self.args:
            name_width = max([(len(arg.name) if arg.name else 0) for arg in self.args])
            for arg in self.args:
                arg_doc_string = arg.print(name_width, prefix)
                ret += f"\n{arg_doc_string}"
        return ret

    def is_same_route_as(self, other: "MethodSpec") -> bool:
        """Two methods are the same route if they share a path and verb."""
        return self.path == other.path and self.verb == other.verb

    def get_bound_function(self, service_instance: Optional[Any]) -> Optional[Callable[..., Any]]:
        """Get the bound method described by this spec.

        The `func_binding`, if a string, resolves to a function on the provided Invocable. Else is just a function."""

        if not self.func_binding:
            logging.error(
                f"MethodSpec attempted to get bound function but func_binding was None. {self}"
            )
            return None

        if isinstance(self.func_binding, str):
            # It's a string; we should resolve against the invocable.
            if not service_instance:
                logging.error(
                    f"MethodSpec attempted to get bound function named {self.func_binding}. "
                    f"But provided service_instance was None. {self}"
                )
                return None

            if not hasattr(service_instance, self.func_binding):
                logging.error(
                    f"MethodSpec attempted to get bound function named {self.func_binding}. "
                    f"But provided service_instance did not have that attribute. {self}"
                )
                return None

            if not callable(getattr(service_instance, self.func_binding)):
                logging.error(
                    f"MethodSpec attempted to get bound function named {self.func_binding}. "
                    f"But that attribute on provided service_instance was not callable. {self}"
                )
                return None

            return getattr(service_instance, self.func_binding)

        elif callable(self.func_binding):
            return self.func_binding

        logging.error(
            f"MethodSpec attempted to get bound function. "
            f"But the func_binding was of type {type(self.func_binding)} and could not be handled. {self}"
        )
        return None


class PackageSpec(CamelModel):
    """A package, representing a remotely instantiable service."""

    # The name of the package
    name: str

    # The docstring of the package
    doc: Optional[str] = None

    # The SDK version this package is deployed with
    sdk_version: str = steamship.__version__

    # Quick O(1) lookup into VERB+NAME
    method_mappings: Dict[str, Dict[str, MethodSpec]] = Field(None, exclude=True, repr=False)

    # TODO: If we upgrade to Pydantic 2xx, we can use @computed_field to include this in dict()
    @property
    def all_methods(self) -> List[MethodSpec]:
        """Return a list of all methods mapped in this Package."""
        if not self.method_mappings:
            return []

        ret = []
        for verb in self.method_mappings:
            for name in self.method_mappings[verb]:
                ret.append(self.method_mappings[verb][name])

        # Sort by name and verb to ease testing
        ret = sorted(ret, key=lambda m: (m.path, m.verb))

        return ret

    def pprint(self, prefix: str = "  ") -> str:
        """Returns a pretty printable representation of this package."""
        underline = "=" * len(self.name)
        ret = f"{self.name}\n{underline}\n"
        if self.doc:
            ret += f"{self.doc}\n\n"
        else:
            ret += "\n"

        methods = self.all_methods

        if methods:
            name_width = max([len(method.path) or 0 for method in methods])
            for method in methods:
                method_doc_string = method.pprint(name_width, prefix)
                ret += f"\n{method_doc_string}"
        return ret

    def import_parent_methods(self, parent: Optional["PackageSpec"] = None):
        if not parent:
            return
        for method in parent.all_methods:
            self.add_method(method.clone(), permit_overwrite_of_existing=True)

    def add_method(self, new_method: MethodSpec, permit_overwrite_of_existing: bool = False):
        """Add a method to the MethodSpec, overwriting the existing if it exists."""
        if not self.method_mappings:
            self.method_mappings = {}

        if new_method.verb not in self.method_mappings:
            self.method_mappings[new_method.verb] = {}

        if (
            new_method.path in self.method_mappings[new_method.verb]
            and not permit_overwrite_of_existing
        ):
            raise SteamshipError(
                message="Attempted to double-register route without explicitly permitting double-registry. "
                "Please include the kwarg permit_overwrite_of_existing=True to confirm your intent. "
                f"Route: {new_method}"
            )

        self.method_mappings[new_method.verb][new_method.path] = new_method

    def get_method(self, http_verb: str, http_path: str) -> Optional[MethodSpec]:
        """Matches the provided HTTP Verb and Path to registered methods.

        This is intended to be the single place where a provided (VERB, PATH) is mapped to a MethodSpec, such
        that if we eventually support path variables (/posts/:id/raw), it can be done within this function.
        """
        verb = Verb(http_verb.strip().upper())
        path = MethodSpec.clean_path(http_path)

        if not self.method_mappings:
            logging.error("PackageSpec.get_method: method_mappings is None.")
            return None

        if verb not in self.method_mappings:
            logging.error(f"PackageSpec.match_route: Verb '{verb}' not found in method_mappings.")
            return None

        if path not in self.method_mappings[verb]:
            logging.error(
                f"PackageSpec.match_route: Path '{path}' not found in method_mappings[{verb}]."
            )
            return None

        return self.method_mappings[verb][path]

    def dict(self, **kwargs) -> dict:
        """Return the dict representation of this object.

        Manually adds the `methods` computed field. Note that if we upgrade to Pydantic 2xx we can automatically
        include this via decorators.
        """
        ret = super().dict(**kwargs)
        ret["methods"] = [m.dict(**kwargs) for m in self.all_methods]
        return ret

    def clone(self) -> "PackageSpec":
        """Return a copy-by-value clone of this PackageSpec."""
        ret = PackageSpec(
            name=deepcopy(self.name), doc=deepcopy(self.doc), sdk_version=deepcopy(self.sdk_version)
        )
        for method in self.all_methods:
            ret.add_method(method.clone())
        return ret



import io
import json
from datetime import datetime

import pytest
import requests
from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin

from steamship import MimeTypes, SteamshipError
from steamship.base.request import SortOrder
from steamship.client import Steamship
from steamship.data import TagKind
from steamship.data.block import Block
from steamship.data.file import File
from steamship.data.tags.tag import Tag


@pytest.mark.usefixtures("client")
def test_file_upload(client: Steamship):
    a = File.create(client=client, handle="foo", content="A", mime_type=MimeTypes.MKD)
    assert a.id is not None
    assert a.mime_type == MimeTypes.MKD
    assert a.handle == "foo"

    b = File.create(client=client, content="B", mime_type=MimeTypes.TXT)
    assert b.id is not None
    assert b.mime_type == MimeTypes.TXT

    assert a.id != b.id

    assert a.raw().decode("utf-8") == "A"
    a.delete()
    b.delete()


@pytest.mark.usefixtures("client")
def test_only_blocks_or_content(client: Steamship):
    with pytest.raises(SteamshipError):
        _ = File.create(client=client, content="A", blocks=[], mime_type=MimeTypes.MKD)


@pytest.mark.usefixtures("client")
def test_file_upload_with_content_and_tags(client: Steamship):
    a = File.create(
        client=client,
        content="ABC",
        mime_type=MimeTypes.MKD,
        tags=[Tag(kind="SomeKind")],
    )
    assert a.id is not None
    assert a.mime_type == MimeTypes.MKD
    assert len(a.tags) == 1
    assert a.tags[0].file_id == a.id
    assert a.tags[0].kind == "SomeKind"
    assert a.raw().decode("utf-8") == "ABC"


def test_file_import_response_dict():
    resp = File.CreateResponse(_bytes=b"some bytes", mime_type=MimeTypes.BINARY)
    to_dict = resp.dict(include={"data_", "mime_type"})
    file_create_response = File.CreateResponse.parse_obj(to_dict)
    assert resp.data_ == file_create_response.data_
    assert resp.mime_type == file_create_response.mime_type


def test_file_import_response_bytes_serialization():
    file_resp = File.CreateResponse(_bytes=b"some bytes", mime_type=MimeTypes.BINARY)
    to_dict = file_resp.dict()
    as_json_string = json.dumps(to_dict)
    as_dict_again = json.loads(as_json_string)
    assert as_dict_again == to_dict


def test_file_upload_with_blocks(client: Steamship):
    a = File.create(
        client=client,
        blocks=[
            Block(text="A", tags=[Tag(kind="BlockTag")]),
            Block(text="B", tags=[Tag(kind="BlockTag")]),
        ],
    )
    assert a.id is not None

    query_result = Block.query(client, f'file_id "{a.id}"')

    def check_blocks(block_list):
        assert len(block_list) == 2
        assert block_list[0].tags is not None
        assert len(block_list[0].tags) == 1
        assert block_list[0].tags[0].kind == "BlockTag"
        assert block_list[0].text == "A"

    assert query_result.blocks is not None
    check_blocks(query_result.blocks)

    # Let's get the file fresh
    aa = File.get(client, _id=a.id)
    check_blocks(aa.blocks)
    a.delete()


def test_file_upload_with_blocks_and_tags(client: Steamship):
    a = File.create(
        client=client,
        blocks=[
            Block(text="A", tags=[Tag(kind="BlockTag")]),
            Block(text="B", tags=[Tag(kind="BlockTag")]),
        ],
        tags=[Tag(kind="FileTag")],
    )
    assert a.id is not None

    blocks = Block.query(client, f'file_id "{a.id}"')

    def check_blocks(block_list):
        assert len(block_list) == 2
        assert block_list[0].tags is not None
        assert len(block_list[0].tags) == 1
        assert block_list[0].tags[0].kind == "BlockTag"
        assert block_list[0].text == "A"

    assert blocks.blocks is not None
    check_blocks(blocks.blocks)

    # Let's get the file fresh
    aa = File.get(client, _id=a.id)
    check_blocks(aa.blocks)
    assert aa.tags is not None
    assert len(aa.tags) == 1
    assert aa.tags[0].kind == "FileTag"

    a.delete()


def test_file_upload_with_tags(client: Steamship):
    a = File.create(
        client=client,
        tags=[Tag(kind="FileTag")],
    )
    assert a.id is not None

    query_result = Tag.query(client, f'filetag and file_id "{a.id}"')

    def check_tags(file_tag_list):
        assert len(file_tag_list) == 1
        assert file_tag_list[0] is not None
        assert file_tag_list[0].kind == "FileTag"

    assert query_result.tags is not None
    check_tags(query_result.tags)

    # Let's get the file fresh
    aa = File.get(client, _id=a.id)
    check_tags(aa.tags)

    a.delete()


def test_query(client: Steamship):
    a = File.create(
        client=client,
        blocks=[
            Block(text="A", tags=[Tag(kind="BlockTag")]),
            Block(text="B"),
        ],
    )
    assert a.id is not None
    b = File.create(
        client=client,
        blocks=[Block(text="A"), Block(text="B")],
        tags=[Tag(kind="FileTag")],
    )
    assert b.id is not None

    files = File.query(client=client, tag_filter_query='blocktag and kind "BlockTag"').files
    assert len(files) == 1
    assert files[0].id == a.id

    files = File.query(client=client, tag_filter_query='filetag and kind "FileTag"').files
    assert len(files) == 1
    assert files[0].id == b.id

    # Test serialization; This shouldn't throw
    out_file = File.get(client, _id=b.id)
    json.dumps(out_file.dict())

    a.delete()
    b.delete()


def test_file_list(client: Steamship):
    a = File.create(
        client=client,
        tags=[Tag(kind="FileTag")],
    )
    b = File.create(
        client=client,
        tags=[Tag(kind="FileTag")],
    )
    c = File.create(
        client=client,
        tags=[Tag(kind="FileTag")],
    )

    files = File.list(client=client).files
    assert len(files) == 3
    assert a in files
    assert b in files
    assert c in files

    a.delete()
    b.delete()
    c.delete()


def test_file_list_paging(client: Steamship):
    a = File.create(
        client=client,
        tags=[Tag(kind="FileTag", value={"name": "A"})],
    )
    b = File.create(
        client=client,
        tags=[Tag(kind="FileTag", value={"name": "B"})],
    )
    c = File.create(
        client=client,
        tags=[Tag(kind="FileTag", value={"name": "C"})],
    )

    page_size = 1
    resp = File.list(client=client, page_size=page_size)
    assert len(resp.files) == page_size
    assert resp.next_page_token is not None
    assert c == resp.files[0]  # default is DESC (createdAt)

    resp = File.list(client=client, page_size=page_size, page_token=resp.next_page_token)
    assert len(resp.files) == page_size
    assert resp.next_page_token is not None
    assert b == resp.files[0]

    resp = File.list(client=client, page_size=page_size, page_token=resp.next_page_token)
    assert len(resp.files) == page_size
    assert resp.next_page_token is None
    assert a == resp.files[0]

    files = File.list(client=client, page_size=page_size, sort_order=SortOrder.ASC).files
    assert len(files) == 1
    assert files[0] == a

    with pytest.raises(SteamshipError):
        File.list(client=client, page_size=-1)

    resp = File.list(client=client, page_token="not-found-foo")  # noqa: S106
    assert len(resp.files) == 3

    a.delete()
    b.delete()
    c.delete()


def test_file_refresh(client: Steamship):
    blockifier_path = PLUGINS_PATH / "blockifiers" / "blockifier.py"
    with deploy_plugin(client, blockifier_path, "blockifier") as (
        plugin,
        version,
        instance,
    ):
        file = File.create(client=client, content="This is a test.")
        assert len(file.blocks) == 0
        file.blockify(plugin_instance=instance.handle).wait()
        file.refresh()  # don't reassign file
        assert len(file.blocks) == 4
        file.delete()


@pytest.mark.usefixtures("client")
def test_append_indices(client: Steamship):
    file = File.create(client, blocks=[Block(text="first")])
    assert len(file.blocks) == 1
    assert file.blocks[0].index_in_file == 0

    appended_block = file.append_block(text="second")
    assert appended_block.index_in_file == 1
    assert len(file.blocks) == 2

    file.refresh()
    assert len(file.blocks) == 2
    assert file.blocks[0].index_in_file == 0
    assert file.blocks[0].text == "first"
    assert file.blocks[1].index_in_file == 1
    assert file.blocks[1].text == "second"


@pytest.mark.usefixtures("client")
def test_file_upload_content_with_tags_and_tag_value(client: Steamship):
    """This test created to test client-side that the multipart file content + tag parsing works correctly"""
    a = File.create(
        client=client,
        handle="foo",
        content="f",
        tags=[Tag(kind="created_at", value={"date-value": datetime.now().isoformat()})],
    )
    assert a.id is not None
    assert a.handle == "foo"
    assert len(a.tags) == 1
    assert a.tags[0].kind == "created_at"
    assert a.tags[0].value["date-value"] is not None

    assert a.raw().decode("utf-8") == "f"
    a.delete()


@pytest.mark.usefixtures("client")
def test_file_create_from_resp(client: Steamship):
    class FakeSession:
        def post(self, url: str, **kwargs):
            resp = requests.Response()
            resp.status_code = 200
            resp.headers = {"Content-Type": "application/text"}
            resp.raw = io.BytesIO(bytes("<html>you done goofed</html>", "utf-8"))
            return resp

    client._session = FakeSession()

    with pytest.raises(SteamshipError) as err:
        # the content here is not relevant to the test
        # this should raise a SteamshipError complaining about expectations.
        File.create(
            client=client, blocks=[Block(text="/usr/bin/python root.py")], mime_type=MimeTypes.TXT
        )

    assert "data does not match expected" in str(err)


@pytest.mark.usefixtures("client")
def test_append_with_tag(client: Steamship):
    file = File.create(client, blocks=[Block(text="first")])
    assert len(file.blocks) == 1
    assert file.blocks[0].index_in_file == 0

    appended_block = file.append_block(text="second", tags=[Tag(kind=TagKind.DOCUMENT)])
    assert appended_block.index_in_file == 1
    assert len(file.blocks) == 2

    file.refresh()
    assert len(file.blocks) == 2
    assert file.blocks[0].index_in_file == 0
    assert file.blocks[0].text == "first"
    assert file.blocks[1].index_in_file == 1
    assert file.blocks[1].text == "second"
    assert len(file.blocks[1].tags) == 1
    assert file.blocks[1].tags[0].kind == TagKind.DOCUMENT


@pytest.mark.usefixtures("client")
def test_file_public_data(client: Steamship):
    file = File.create(client, content=bytes("This is a test", "utf-8"), public_data=True)
    assert file.public_data

    # Intentionally no API key
    response = requests.get(file.raw_data_url)

    assert response.text == "This is a test"
    assert response.headers["content-type"] == MimeTypes.TXT


@pytest.mark.usefixtures("client")
def test_file_private_data(client: Steamship):
    file = File.create(client, content=bytes("This is a test", "utf-8"))
    assert not file.public_data

    # Intentionally no API key
    failed_response = requests.get(file.raw_data_url)
    assert not failed_response.ok

    # Should still be able to get with API key
    response = requests.get(
        file.raw_data_url,
        headers={"Authorization": f"Bearer {client.config.api_key.get_secret_value()}"},
    )

    assert response.text == "This is a test"
    assert response.headers["content-type"] == MimeTypes.TXT


@pytest.mark.usefixtures("client")
def test_set_public_data(client: Steamship):
    file = File.create(client, content=bytes("This is a test", "utf-8"))

    assert not file.public_data

    # With public data false, call should fail
    # Intentionally no API key
    failed_response = requests.get(file.raw_data_url)
    assert not failed_response.ok

    # With public data true, call should succeed
    file.set_public_data(True)
    # Intentionally no API key
    response = requests.get(file.raw_data_url)
    assert response.text == "This is a test"
    assert response.headers["content-type"] == MimeTypes.TXT

    # Setting back to false, should fail again
    file.set_public_data(False)
    # Intentionally no API key
    failed_response = requests.get(file.raw_data_url)
    assert not failed_response.ok



"""Answers questions with the assistance of a VectorSearch plugin."""
from typing import Any, List, Optional, Union, cast

from steamship import Block, Steamship, Tag, Task
from steamship.agents.llms import OpenAI
from steamship.agents.schema import AgentContext, Tool
from steamship.agents.utils import get_llm, with_llm
from steamship.data.plugin.index_plugin_instance import EmbeddingIndexPluginInstance
from steamship.utils.repl import ToolREPL

DEFAULT_QUESTION_ANSWERING_PROMPT = (
    "Use the following pieces of memory to answer the question at the end. "
    """If you don't know the answer, just say that you don't know, don't try to make up an answer.

{source_text}

Question: {question}

Helpful Answer:"""
)


DEFAULT_SOURCE_DOCUMENT_PROMPT = "Source Document: {text}"


class VectorSearchQATool(Tool):
    """Tool to answer questions with the assistance of a vector search plugin."""

    name: str = "VectorSearchQATool"
    human_description: str = "Answers questions with help from a Vector Database."
    agent_description: str = (
        "Used to answer questions. ",
        "The input should be a plain text question. ",
        "The output is a plain text answer",
    )
    embedding_index_handle: Optional[str] = "embedding-index"
    embedding_index_version: Optional[str] = None
    question_answering_prompt: Optional[str] = DEFAULT_QUESTION_ANSWERING_PROMPT
    source_document_prompt: Optional[str] = DEFAULT_SOURCE_DOCUMENT_PROMPT
    embedding_index_config: Optional[dict] = {
        "embedder": {
            "plugin_handle": "openai-embedder",
            "plugin_instance-handle": "text-embedding-ada-002",
            "fetch_if_exists": True,
            "config": {"model": "text-embedding-ada-002", "dimensionality": 1536},
        }
    }
    load_docs_count: int = 2
    embedding_index_instance_handle: str = "default-embedding-index"

    def get_embedding_index(self, client: Steamship) -> EmbeddingIndexPluginInstance:
        index = client.use_plugin(
            plugin_handle=self.embedding_index_handle,
            instance_handle=self.embedding_index_instance_handle,
            config=self.embedding_index_config,
            fetch_if_exists=True,
        )
        return cast(EmbeddingIndexPluginInstance, index)

    def answer_question(self, question: str, context: AgentContext) -> List[Block]:
        index = self.get_embedding_index(context.client)
        task = index.search(question, k=self.load_docs_count)
        task.wait()

        source_texts = []

        for item in task.output.items:
            if item.tag and item.tag.text:
                item_data = {"text": item.tag.text}
                source_texts.append(self.source_document_prompt.format(**item_data))

        final_prompt = self.question_answering_prompt.format(
            **{"source_text": "\n".join(source_texts), "question": question}
        )

        return get_llm(context).complete(prompt=final_prompt)

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        """Answers questions with the assistance of an Embedding Index plugin.

        Inputs
        ------
        input: List[Block]
            A list of blocks to be rewritten if text-containing.
        memory: AgentContext
            The active AgentContext.

        Output
        ------
        output: List[Blocks]
            A lit of blocks containing the answers.
        """

        output = []
        for input_block in tool_input:
            if not input_block.is_text():
                continue
            for output_block in self.answer_question(input_block.text, context):
                output.append(output_block)
        return output


if __name__ == "__main__":
    tool = VectorSearchQATool()
    repl = ToolREPL(tool)

    with repl.temporary_workspace() as client:
        index = tool.get_embedding_index(client)
        index.insert([Tag(text="Ted loves apple pie."), Tag(text="The secret passcode is 1234.")])
        repl.run_with_client(
            client, context=with_llm(context=AgentContext(), llm=OpenAI(client=client))
        )



from __future__ import annotations

from typing import List

from steamship.data.file import Block
from steamship.plugin.outputs.plugin_output import PluginOutput


class RawBlockAndTagPluginOutput(PluginOutput):
    blocks: List[Block]



from __future__ import annotations

import time
from typing import Any, Dict, Optional, Type

from pydantic import BaseModel, Field

from steamship import SteamshipError
from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import DeleteRequest, IdentifierRequest, Request
from steamship.data.invocable_init_status import InvocableInitStatus
from steamship.data.workspace import Workspace
from steamship.utils.url import Verb


class CreatePackageInstanceRequest(Request):
    id: str = None
    package_id: str = None
    package_handle: str = None
    package_version_id: str = None
    package_version_handle: str = None
    handle: str = None
    fetch_if_exists: bool = None
    config: Dict[str, Any] = None
    workspace_id: str = None


class PackageInstance(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    handle: str = None
    package_id: str = None
    package_handle: Optional[str] = None
    user_handle: str = None
    package_version_id: str = None
    package_version_handle: Optional[str] = None
    user_id: str = None
    invocation_url: str = None
    config: Dict[str, Any] = None
    workspace_id: str = None
    workspace_handle: str = None
    init_status: Optional[InvocableInitStatus] = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["packageInstance"] if "packageInstance" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        package_id: str = None,
        package_handle: str = None,
        package_version_id: str = None,
        package_version_handle: str = None,
        handle: str = None,
        fetch_if_exists: bool = None,
        config: Dict[str, Any] = None,
    ) -> PackageInstance:
        req = CreatePackageInstanceRequest(
            handle=handle,
            package_id=package_id,
            package_handle=package_handle,
            package_version_id=package_version_id,
            package_version_handle=package_version_handle,
            fetch_if_exists=fetch_if_exists,
            config=config,
        )

        return client.post("package/instance/create", payload=req, expect=PackageInstance)

    def delete(self) -> PackageInstance:
        req = DeleteRequest(id=self.id)
        return self.client.post("package/instance/delete", payload=req, expect=PackageInstance)

    def load_missing_workspace_handle(self):
        if (
            self.client is not None
            and self.workspace_handle is None
            and self.workspace_id is not None
        ):
            # Get the workspaceHandle
            workspace = Workspace.get(self.client, id_=self.workspace_id)
            if workspace:
                self.workspace_handle = workspace.handle

    @staticmethod
    def get(client: Client, handle: str) -> PackageInstance:
        return client.post(
            "package/instance/get", IdentifierRequest(handle=handle), expect=PackageInstance
        )

    def invoke(
        self, path: str, verb: Verb = Verb.POST, timeout_s: Optional[float] = None, **kwargs
    ):
        self.load_missing_workspace_handle()
        if path[0] == "/":
            path = path[1:]

        return self.client.call(
            verb=verb,
            operation=f"/{self.workspace_handle or '_'}/{self.handle or '_'}/{path}",
            payload=kwargs,
            is_package_call=True,
            package_owner=self.user_handle,
            package_id=self.package_id,
            package_instance_id=self.id,
            as_background_task=False,
            timeout_s=timeout_s,
        )

    def full_url_for(self, path: str):
        return f"{self.invocation_url}{path}"

    def refresh_init_status(self):
        new_self = PackageInstance.get(self.client, handle=self.handle)
        self.init_status = new_self.init_status

    def wait_for_init(
        self,
        max_timeout_s: float = 180,
        retry_delay_s: float = 1,
    ):
        """Polls and blocks until the init has succeeded or failed (or timeout reached).

        Parameters
        ----------
        max_timeout_s : int
            Max timeout in seconds. Default: 180s. After this timeout, an exception will be thrown.
        retry_delay_s : float
            Delay between status checks. Default: 1s.
        """
        t0 = time.perf_counter()
        while (
            time.perf_counter() - t0 < max_timeout_s
            and self.init_status == InvocableInitStatus.INITIALIZING
        ):
            time.sleep(retry_delay_s)
            self.refresh_init_status()

        # If the task did not complete within the timeout, throw an error
        if self.init_status == InvocableInitStatus.INITIALIZING:
            raise SteamshipError(
                message=f"Package Instance {self.id} did not complete within requested timeout of {max_timeout_s}s. The init is still running on the server. You can retrieve its status via PackageInstance.get() or try waiting again with wait_for_init()."
            )



from enum import Enum


class MimeTypes(str, Enum):
    UNKNOWN = "unknown"
    TXT = "text/plain"
    JSON = "application/json"
    MKD = "text/markdown"
    EPUB = "application/epub+zip"
    PDF = "application/pdf"
    JPG = "image/jpeg"
    PNG = "image/png"
    TIFF = "image/tiff"
    GIF = "image/gif"
    HTML = "text/html"
    DOC = "application/msword"
    DOCX = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    PPT = "applicatino/ms-powerpoint"
    PPTX = "application/vnd.openxmlformats-officedocument.presentationml.presentation"
    RTF = "application/rtf"
    BINARY = "application/octet-stream"
    STEAMSHIP_BLOCK_JSON = "application/vnd.steamship-block.json.v1"
    WAV = "audio/wav"
    MP3 = "audio/mp3"
    MP4_VIDEO = "video/mp4"
    MP4_AUDIO = "audio/mp4"
    WEBM_VIDEO = "video/webm"
    WEBM_AUDIO = "audio/webm"
    FILE_JSON = "fileJson"


class ContentEncodings:
    BASE64 = "base64"


TEXT_MIME_TYPES = [
    MimeTypes.TXT,
    MimeTypes.MKD,
    MimeTypes.HTML,
    MimeTypes.DOCX,
    MimeTypes.PPTX,
]



"""Tests that base classes can register routes, but they're overridden by subclasses.

This tests a fix (in the same PR) to a behavior in which each new Subclass of Invocable would register its route
table afresh, overwriting whatever routes were registered by an ancestor.
"""
import pytest

from steamship import Steamship, SteamshipError
from steamship.agents.llms import OpenAI
from steamship.agents.react import ReACTAgent
from steamship.base.package_spec import MethodSpec
from steamship.experimental.package_starters.telegram_agent import TelegramAgentService
from steamship.invocable import Invocable, InvocableRequest, Invocation, post
from steamship.utils.url import Verb


class L1Invocable(Invocable):
    @post("foo")
    def foo(self) -> str:
        return "l1_foo"

    @post("bar")
    def bar(self) -> str:
        return "l1_bar"

    @post("baz")
    def baz(self) -> str:
        return "l1_baz"


class L2Invocable(L1Invocable):
    @post("bar")
    def bar(self) -> str:
        return "l2_bar"

    @post("baz")
    def baz(self) -> str:
        return "l2_baz"


class L2Invocable2(L1Invocable):
    @post("bar")
    def bar2(self) -> str:
        return "l22_bar"

    @post("baz")
    def baz2(self) -> str:
        return "l22_baz"


class L3Invocable(L2Invocable):
    @post("baz")
    def baz(self) -> str:
        return "l3_baz"


class L3Invocable2(L2Invocable2):
    @post("baz")
    def baz3(self) -> str:
        return "l32_baz"


class L4Invocable(L2Invocable):
    """Test registering a lambda route"""

    def __init__(
        self, *args, update: bool = False, permit_overwrite_of_existing: bool = False, **kwargs
    ):
        super().__init__(*args, **kwargs)

        if update:

            def inline_func():
                return "Lambda"

            method_spec = MethodSpec(
                name="/baz", verb=Verb.POST, func_binding=inline_func, returns="str"
            )
            self.add_api_route(
                method_spec, permit_overwrite_of_existing=permit_overwrite_of_existing
            )


class MyAgentService(TelegramAgentService):
    pass


def invoke(o: Invocable, path: str):
    req = InvocableRequest(invocation=Invocation(http_verb="POST", invocation_path=path))
    return o(req)


def test_l1_routes():
    """Tests that we can inspect the L1 routes"""
    l1 = L1Invocable()
    assert l1._package_spec.method_mappings[Verb.POST]["/foo"].func_binding == "foo"
    assert l1._package_spec.method_mappings[Verb.POST]["/bar"].func_binding == "bar"
    assert l1._package_spec.method_mappings[Verb.POST]["/baz"].func_binding == "baz"
    assert invoke(l1, "foo") == "l1_foo"
    assert invoke(l1, "bar") == "l1_bar"
    assert invoke(l1, "baz") == "l1_baz"

    routes = [m["path"] for m in l1.__steamship_dir__()["methods"]]
    assert "/foo" in routes
    assert "/bar" in routes
    assert "/baz" in routes
    assert len(routes) == 6  # __instance__init + 2x __dir__


def test_l2_routes():
    """Tests that we can inspect the L1 routes"""
    l2 = L2Invocable()
    assert l2._package_spec.method_mappings[Verb.POST]["/foo"].func_binding == "foo"
    assert l2._package_spec.method_mappings[Verb.POST]["/bar"].func_binding == "bar"
    assert l2._package_spec.method_mappings[Verb.POST]["/baz"].func_binding == "baz"
    assert invoke(l2, "foo") == "l1_foo"
    assert invoke(l2, "bar") == "l2_bar"
    assert invoke(l2, "baz") == "l2_baz"

    routes = [m["path"] for m in l2.__steamship_dir__()["methods"]]
    assert len(routes) == 6  # __instance__init + 2x __dir__
    assert "/foo" in routes
    assert "/bar" in routes
    assert "/baz" in routes


def test_l3_routes():
    """Tests that we can inspect the L1 routes"""
    l3 = L3Invocable()
    assert l3._package_spec.method_mappings[Verb.POST]["/foo"].func_binding == "foo"
    assert l3._package_spec.method_mappings[Verb.POST]["/bar"].func_binding == "bar"
    assert l3._package_spec.method_mappings[Verb.POST]["/baz"].func_binding == "baz"
    assert invoke(l3, "foo") == "l1_foo"
    assert invoke(l3, "bar") == "l2_bar"
    assert invoke(l3, "baz") == "l3_baz"

    routes = [m["path"] for m in l3.__steamship_dir__()["methods"]]
    assert len(routes) == 6  # __instance__init + 2x __dir__
    assert "/foo" in routes
    assert "/bar" in routes
    assert "/baz" in routes


def test_l22_routes():
    """Tests that we can inspect the L1 routes"""
    l22 = L2Invocable2()
    assert l22._package_spec.method_mappings[Verb.POST]["/foo"].func_binding == "foo"
    assert l22._package_spec.method_mappings[Verb.POST]["/bar"].func_binding == "bar2"
    assert l22._package_spec.method_mappings[Verb.POST]["/baz"].func_binding == "baz2"
    assert invoke(l22, "foo") == "l1_foo"
    assert invoke(l22, "bar") == "l22_bar"
    assert invoke(l22, "baz") == "l22_baz"

    routes = [m["path"] for m in l22.__steamship_dir__()["methods"]]
    assert len(routes) == 6  # __instance__init + 2x __dir__
    assert "/foo" in routes
    assert "/bar" in routes
    assert "/baz" in routes


def test_l32_routes():
    """Tests that we can inspect the L1 routes"""
    l32 = L3Invocable2()

    routes = [m["path"] for m in l32.__steamship_dir__()["methods"]]
    assert len(routes) == 6  # __instance__init + 2x __dir__
    assert "/foo" in routes
    assert "/bar" in routes
    assert "/baz" in routes

    assert l32._package_spec.method_mappings[Verb.POST]["/foo"].func_binding == "foo"
    assert l32._package_spec.method_mappings[Verb.POST]["/bar"].func_binding == "bar2"
    assert l32._package_spec.method_mappings[Verb.POST]["/baz"].func_binding == "baz3"
    assert invoke(l32, "foo") == "l1_foo"
    assert invoke(l32, "bar") == "l22_bar"
    assert invoke(l32, "baz") == "l32_baz"


def test_l4_routes():
    """Tests that we can inspect the L1 routes"""
    l4 = L4Invocable(update=True, permit_overwrite_of_existing=True)

    routes = [m["path"] for m in l4.__steamship_dir__()["methods"]]
    assert len(routes) == 6  # __instance__init + 2x __dir__
    assert "/foo" in routes
    assert "/bar" in routes
    assert "/baz" in routes

    assert l4._package_spec.method_mappings[Verb.POST]["/foo"].func_binding == "foo"
    assert l4._package_spec.method_mappings[Verb.POST]["/bar"].func_binding == "bar"
    assert callable(l4._package_spec.method_mappings[Verb.POST]["/baz"].func_binding)
    assert invoke(l4, "foo") == "l1_foo"
    assert invoke(l4, "bar") == "l2_bar"
    assert invoke(l4, "baz") == "Lambda"

    # Make sure that we didn't overwrite anything else
    l42 = L4Invocable(update=False, permit_overwrite_of_existing=True)
    assert invoke(l4, "baz") == "Lambda"
    assert invoke(l42, "baz") == "l2_baz"

    # Without permitting overwrite, this throws an exception
    with pytest.raises(SteamshipError):
        L4Invocable(update=True, permit_overwrite_of_existing=False)


@pytest.mark.usefixtures("client")
def test_telegram_agent(client: Steamship):
    a = MyAgentService(
        client=client,
        config={"botToken": "foo"},
        incoming_message_agent=ReACTAgent(tools=[], llm=OpenAI(client=client)),
    )
    assert a._package_spec.method_mappings[Verb.POST]["/answer"].func_binding == "answer"
    routes = [m["path"] for m in a.__steamship_dir__()["methods"]]
    assert "/answer" in routes
    assert "/respond" in routes
    assert "/webhook_info" in routes
    assert "/info" in routes



from .assembly_speech_to_text_tool import AssemblySpeechToTextTool
from .fetch_audio_urls_from_rss_tool import FetchAudioUrlsFromRssTool
from .whisper_speech_to_text_tool import WhisperSpeechToTextTool

__all__ = [
    "AssemblySpeechToTextTool",
    "FetchAudioUrlsFromRssTool",
    "WhisperSpeechToTextTool",
]



"""Tool for generating images."""
from steamship.agents.tools import AudioGeneratorTool
from steamship.utils.repl import ToolREPL


class GenerateSpeechTool(AudioGeneratorTool):
    """Tool to generate audio from text."""

    name: str = "GenerateSpokenAudio"
    human_description: str = "Generates spoken audio from text."
    agent_description: str = (
        "Used to generate spoken audio from text prompts. Only use if the user has asked directly for a "
        "an audio version of output. When using this tool, the input should be a plain text string containing the "
        "content to be spoken."
    )
    generator_plugin_handle: str = "elevenlabs"


if __name__ == "__main__":
    ToolREPL(GenerateSpeechTool()).run()



from __future__ import annotations

import json
from json import JSONDecodeError
from typing import Any

from steamship.base.model import CamelModel


class Hit(CamelModel):
    id: str = None
    index: int = None
    index_source: str = None
    value: str = None
    score: float = None
    external_id: str = None
    external_type: str = None
    metadata: Any = None
    query: str = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        metadata = kwargs.get("metadata")
        if metadata is not None:
            try:
                self.metadata = json.loads(metadata)
            except JSONDecodeError:
                pass



from typing import Any, List, Union

from steamship import Block, Task
from steamship.agents.schema import AgentContext, Tool
from steamship.utils.repl import ToolREPL


class KnockKnockTool(Tool):
    """
    Example tool to illustrate how one might initiate the beginning of a joke.

    Experimentally, the conversational LLM which underlies the agent ought to take over from there on out and
    be capable of completing the joke once it's begun.
    """

    name = "KnockKnockTool"
    human_description = "Starts a Knock-Knock Joke."
    agent_description = "Used to begin telling a Knock Knock joke."

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        return [Block(text="Knock-Knock...")]


if __name__ == "__main__":
    ToolREPL(KnockKnockTool()).run()



from steamship.invocable import Invocable, InvocableResponse, get, post
from steamship.invocable.lambda_handler import create_safe_handler


class HelloWorld(Invocable):
    @post("greet")
    def greet(self, name: str = "Person") -> InvocableResponse:
        return InvocableResponse(string=f"Hello, {name}")

    @get("workspace")
    def workspace(self) -> InvocableResponse:
        return InvocableResponse(string=self.client.config.workspace_id)


handler = create_safe_handler(known_invocable_for_testing=HelloWorld)



import logging
from logging import StreamHandler
from typing import cast

from fluent.handler import FluentRecordFormatter

from steamship.agents.logging import AgentLogging

LOGGING_FORMAT = {
    "level": "%(levelname)s",
    "host": "%(hostname)s",
    "where": "%(module)s.%(filename)s.%(funcName)s:%(lineno)s",
    "type": "%(levelname)s",
    "stack_trace": "%(exc_text)s",
    "message_type": "%(message_type)s",
    "component_name": "%(component_name)s",
    AgentLogging.IS_MESSAGE: f"%({AgentLogging.IS_MESSAGE})s",  # b doesn't work. Unsure how to make a bool
    AgentLogging.AGENT_NAME: f"%({AgentLogging.AGENT_NAME})s",
    AgentLogging.MESSAGE_AUTHOR: f"%({AgentLogging.MESSAGE_AUTHOR})s",
    AgentLogging.MESSAGE_TYPE: f"%({AgentLogging.MESSAGE_TYPE})s",
    AgentLogging.TOOL_NAME: f"%({AgentLogging.TOOL_NAME})s",
}


class DevelopmentLoggingHandler(StreamHandler):
    """A logging handler for developing Steamship Agents, Tools, Packages, and Plugins locally."""

    log_level: any
    log_level_with_message_type: any

    def __init__(self, log_level: any = logging.WARN, log_level_for_messages: any = logging.INFO):
        StreamHandler.__init__(self)
        formatter = FluentRecordFormatter(LOGGING_FORMAT, fill_missing_fmt_key=True)
        self.setFormatter(formatter)
        self.log_level = log_level
        self.log_level_for_messages = log_level_for_messages

    def _emit_regular(self, message_dict: dict):
        level = message_dict.get("level", None)
        message = message_dict.get("message", None)
        print(f"[{level}] {message}")

    def _emit_message(self, message_dict: dict):
        author = message_dict.get(AgentLogging.MESSAGE_AUTHOR, "Unknown")
        message = message_dict.get("message", None)
        message_type = message_dict.get(AgentLogging.MESSAGE_TYPE, AgentLogging.MESSAGE)

        print(f"[{author} {message_type}] {message}")

    def emit(self, record):
        """Emit the record, printing it to console out.

        We rely on TWO logging levels for the mechanics of this LoggingHandler:

        - One for standard logging
        - One for specific system Agent-related events, flagged with metadata

        This is to permit INFO-level logging of key Agent/Tool actions without committing the user to see all
        INFO-level logging globally.

        A future implementation may use a cascade of loggers attached to the AgentContext to do this more cleanly.
        """
        message_dict = cast(dict, self.format(record))

        # It will be returned as a string representation of a bool
        is_message = message_dict.get(AgentLogging.IS_MESSAGE, None) == "True"

        if record.levelno >= self.log_level and not is_message:
            return self._emit_regular(message_dict)
        elif record.levelno >= self.log_level_for_messages and is_message:
            return self._emit_message(message_dict)



from typing import Optional

from steamship.agents.tools.text_generation.text_rewrite_tool import TextRewritingTool
from steamship.utils.repl import ToolREPL

DEFAULT_PERSONALITY = """A jolly pirate that addresses his friends as 'Matey."""

DEFAULT_PROMPT = """Instructions:
Please rewrite the following passage according to the provided voice, mood, and personality.

Personality:
{personality}

Passage:
{{input}}

Rewritten Passage:"""


class PersonalityTool(TextRewritingTool):
    """
    Example tool to illustrate rewriting a statement according to a particular personality.
    """

    name: str = "PersonalityTool"
    human_description: str = "Rewrites a response with the given personality."
    agent_description: str = "Used to provide a response with a particular personality. Takes a message as input, and provides a message as output."
    personality: str = DEFAULT_PERSONALITY
    rewrite_prompt: str = DEFAULT_PROMPT

    def __init__(
        self, personality: Optional[str] = None, rewrite_prompt: Optional[str] = None, **kwargs
    ):
        _rewrite_prompt = rewrite_prompt or DEFAULT_PROMPT
        kwargs["rewrite_prompt"] = kwargs.get(
            "rewrite_prompt", _rewrite_prompt.format(personality=personality or DEFAULT_PERSONALITY)
        )
        super().__init__(**kwargs)


if __name__ == "__main__":
    tool = PersonalityTool(
        name="BootleggerVibe",
        personality="A 1920s bootlegger, who is always trying to let you in on a little secret.",
    )
    ToolREPL(tool).run()



from typing import List, Optional

from steamship import Block, PluginInstance, Steamship
from steamship.agents.schema import LLM

PLUGIN_HANDLE = "gpt-4"


class OpenAI(LLM):
    """LLM that uses Steamship's OpenAI plugin to generate completions.

    NOTE: By default, this LLM uses the `gpt-3.5-turbo` model. Valid model
    choices are `gpt-3.5-turbo` and `gpt-4`.
    """

    generator: PluginInstance
    client: Steamship

    def __init__(
        self, client, model_name: str = "gpt-3.5-turbo", temperature: float = 0.4, *args, **kwargs
    ):
        """Create a new instance.

        Valid model names are:
         - gpt-4
         - gpt-3.5-turbo
        """
        client = client
        generator = client.use_plugin(
            PLUGIN_HANDLE, config={"model": model_name, "temperature": temperature}
        )
        super().__init__(client=client, generator=generator, *args, **kwargs)

    def complete(self, prompt: str, stop: Optional[str] = None) -> List[Block]:
        options = {}
        if stop:
            options["stop"] = stop
        action_task = self.generator.generate(text=prompt, options=options)
        action_task.wait()
        return action_task.output.blocks



from assets.plugins.taggers.plugin_trainable_tagger import (
    TRAINING_PARAMETERS,
    TestTrainableTaggerPlugin,
    make_train_response,
)
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import File
from steamship.base import Task
from steamship.data.block import Block
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.request import PluginRequestContext

TEST_REQ = BlockAndTagPluginInput(
    file=File(
        blocks=[
            Block(
                id="ABC",
                text="Once upon a time there was a magical ship. "
                "The ship was powered by STEAM. The ship went to the moon.",
            )
        ]
    )
)
TEST_PLUGIN_REQ = PluginRequest(
    data=TEST_REQ, context=PluginRequestContext(plugin_instance_id="000")
)
TEST_PLUGIN_REQ_DICT = TEST_PLUGIN_REQ.dict(by_alias=True)


def _test_resp(res):
    pass
    # assert len(res.data.file.tags) == 3
    # TODO: Finish writing steamship_tests in a future PR once the general design of this is finished.


def test_trainable_tagger():
    client = get_steamship_client()
    assert client is not None

    plugin = TestTrainableTaggerPlugin(client=client)
    assert plugin.client is not None

    # STEP 1. Training Parameters
    # The first part of trainable is to produce trainable parameters. The end-user may offer inputs to this,
    # but ultimately it is the plugin itself which decides upon the final set of trainable parameters.
    tagger1 = plugin.get_training_parameters(
        PluginRequest(
            data=TrainingParameterPluginInput(),
            status=Task(task_id="000"),
            context=PluginRequestContext(plugin_instance_id="000"),
        )
    )
    assert tagger1.data.dict() == TRAINING_PARAMETERS.dict()
    tagger2 = plugin.get_training_parameters_endpoint(
        **PluginRequest(
            data=TrainingParameterPluginInput(),
            status=Task(task_id="000"),
            context=PluginRequestContext(plugin_instance_id="000"),
        ).dict()
    )
    assert tagger2.data.dict() == TRAINING_PARAMETERS.dict()
    assert tagger2.data.training_epochs == TRAINING_PARAMETERS.training_epochs

    # STEP 2. Training
    # The first part of trainable is to produce your own trainable parameters.
    model = plugin.model_cls()()
    tagger1 = plugin.train(
        PluginRequest(
            data=TrainPluginInput(
                plugin_instance="foo", training_params=TRAINING_PARAMETERS.training_params
            ),
            status=Task(task_id="000"),
            context=PluginRequestContext(plugin_instance_id="000"),
        ),
        model,
    )
    train_response = make_train_response()
    train_response.data.archive_path = "000/default.zip"
    train_response.data = train_response.data.dict(by_alias=True)
    assert tagger1.data == train_response.data

    tagger2 = plugin.train_endpoint(
        **PluginRequest(
            data=TrainPluginInput(
                plugin_instance="foo", training_params=TRAINING_PARAMETERS.training_params
            ),
            status=Task(task_id="000"),
            context=PluginRequestContext(plugin_instance_id="000"),
        ).dict()
    )

    assert tagger2.data == train_response.data

    # STEP 3. Run
    res = plugin.run(TEST_PLUGIN_REQ)
    _test_resp(res)

    # The endpoints take a kwargs block which is transformed into the appropriate JSON object
    res2 = plugin.run_endpoint(**TEST_PLUGIN_REQ_DICT)
    _test_resp(res2)



from pathlib import Path
from typing import Type, Union

import pytest

from steamship.base.client import Client
from steamship.invocable import Config, InvocableResponse
from steamship.invocable.plugin_service import PluginRequest, PluginService
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput
from steamship.plugin.outputs.training_parameter_plugin_output import TrainingParameterPluginOutput
from steamship.plugin.tagger import TrainableTagger
from steamship.plugin.trainable_model import TrainableModel


class ValidStringToStringPlugin(PluginService):
    def run(self, request: PluginRequest[str]) -> Union[str, InvocableResponse[str]]:
        pass


class EmptyConfig(Config):
    pass


class ValidTrainableStringToStringModel(TrainableModel[EmptyConfig]):
    def load_from_folder(self, checkpoint_path: Path):
        pass

    def train(self, input: TrainPluginInput) -> TrainPluginOutput:
        pass

    def train_status(self, input: TrainPluginInput) -> TrainPluginOutput:
        pass

    def save_to_folder(self, checkpoint_path: Path):
        pass


class ValidTrainableStringToStringPlugin(TrainableTagger):
    @classmethod
    def config_cls(cls) -> Type[Config]:
        return EmptyConfig

    def model_cls(self) -> Type[TrainableModel]:
        return ValidTrainableStringToStringModel

    def get_steamship_client(self) -> Client:
        return self.client

    def run_with_model(
        self, request: PluginRequest[str], model: TrainableModel
    ) -> Union[str, InvocableResponse[str]]:
        pass

    def get_training_parameters(
        self, request: PluginRequest[TrainingParameterPluginInput]
    ) -> InvocableResponse[TrainingParameterPluginOutput]:
        return InvocableResponse(data=TrainingParameterPluginOutput())

    def train(
        self, request: PluginRequest[TrainPluginInput], model
    ) -> InvocableResponse[TrainPluginOutput]:
        return InvocableResponse(data=TrainPluginOutput())

    def train_status(
        self, request: PluginRequest[TrainPluginInput], model
    ) -> InvocableResponse[TrainPluginOutput]:
        return InvocableResponse(data=TrainPluginOutput())


#
# Tests plugin initialization
# --------------------------------------------


def test_plugin_service_is_abstract():
    with pytest.raises(TypeError):
        PluginService()


def test_plugin_service_must_implement_run_and_subclass_request_from_dict():
    class BadPlugin(PluginService):
        @classmethod
        def config_cls(cls) -> Type[Config]:
            return Config

    with pytest.raises(TypeError):
        BadPlugin()

    ValidStringToStringPlugin()


#
# Tests for the `run` method
# --------------------------------------------


def test_run_succeeds():
    plugin = ValidStringToStringPlugin()
    plugin.run(PluginRequest(data=""))
    # Note: there is no run endpoint implemented automatically, since the path depends on the plugin type.
    # TODO: Should we standardize all plugins to use /run?


#
# Tests for the `get_training_params` method
# --------------------------------------------


def test_plugin_does_not_have_training_param_endpoint():
    plugin = ValidStringToStringPlugin()
    # It has it from the base class
    assert not hasattr(plugin, "get_training_parameters")


def test_with_override_get_training_params_succeeds():
    trainable_plugin = ValidTrainableStringToStringPlugin()
    trainable_plugin.get_training_parameters(PluginRequest(data=TrainingParameterPluginInput()))
    trainable_plugin.get_training_parameters_endpoint()


#
# Tests for the `train` method
# --------------------------------------------


def test_non_trainable_plugin_lacks_train():
    plugin = ValidStringToStringPlugin()
    # It has it from the base class
    assert not hasattr(plugin, "train")


def test_with_override_train_succeeds():
    trainable_plugin = ValidTrainableStringToStringPlugin()
    model = trainable_plugin.model_cls()()
    trainable_plugin.train(PluginRequest(data=TrainPluginInput(plugin_instance="Foo")), model)
    trainable_plugin.train_endpoint()



class AgentLogging:
    """These keys are for use in the `extra` field of agent logging operations. #noqa: RST203

    For now, they are manually applied at the time of logging. In the future, the AgentContext may provide a logger
    which fills some automatically.

    For example:

    logging.info("I should use tool MakeAPicture", extra={
        AgentLogging.AGENT_NAME: self.name,
        AgentLogging.IS_AGENT_MESSAGE: True,
        AgentLogging.MESSAGE_TYPE: AgentLogging.THOUGHT
    }) # noqa: RST203

    This provides:

    * Structured additions to Fluent/Elastic that help with internal debugging.
    * Helpful output in development mode
    * [Eventual] User-visible logs
    * [Eventual] Visualiations about tool execution and ReAct reasoning

    """

    AGENT_NAME = "agent_name"
    TOOL_NAME = "tool_name"
    IS_MESSAGE = "is_message"

    MESSAGE_TYPE = "agent_message_type"
    THOUGHT = "thought"
    OBSERVATION = "observation"
    ACTION = "action"
    MESSAGE = "message"

    MESSAGE_AUTHOR = "message_author"
    USER = "user"
    AGENT = "agent"
    TOOL = "tool"
    SYSTEM = "system"



import json
import logging
import signal
import sys
import time
from os import getenv, path
from typing import Optional

import click

import steamship
from steamship import Steamship, SteamshipError
from steamship.base.configuration import Configuration
from steamship.cli.create_instance import create_instance
from steamship.cli.deploy import (
    PackageDeployer,
    PluginDeployer,
    bundle_deployable,
    update_config_template,
)
from steamship.cli.local_server.server import SteamshipHTTPServer
from steamship.cli.manifest_init_wizard import manifest_init_wizard
from steamship.cli.requirements_init_wizard import requirements_init_wizard
from steamship.cli.ship_spinner import ship_spinner
from steamship.cli.utils import find_api_py, get_api_module
from steamship.data.manifest import DeployableType, Manifest
from steamship.data.user import User
from steamship.invocable.lambda_handler import get_class_from_module


@click.group()
def cli():
    pass


def initialize(suppress_message: bool = False):
    logging.root.setLevel(logging.FATAL)
    if not suppress_message:
        click.echo(f"Steamship Python CLI version {steamship.__version__}")


@click.command()
def login():
    """Log in to Steamship, creating ~/.steamship.json"""
    initialize()
    click.echo("Logging into Steamship.")
    if sys.argv[1] == "login":
        if Configuration.default_config_file_has_api_key():
            overwrite = click.confirm(
                text="You already have an API key in your .steamship.json file.  Do you want to remove it and login?",
                default=False,
            )
            if not overwrite:
                sys.exit(0)
            Configuration.remove_api_key_from_default_config()

        # Carry on with login
        client = Steamship()
        user = User.current(client)
        click.secho(f"🚢🚢🚢 Hooray! You're logged in with user handle: {user.handle} 🚢🚢🚢", fg="green")


@click.command()
def ships():
    """Ship some ships"""
    initialize()
    click.secho("Here are some ships:", fg="cyan")
    with ship_spinner():
        time.sleep(5)
    click.secho()


@click.command()
@click.option(
    "--port",
    "-p",
    type=int,
    default=8080,
    help="Port to host the server on.",
)
@click.option(
    "--invocable_handle",
    "-i",
    type=str,
    default=None,
    help="Handle of the package or plugin being hosted.",
)
@click.option(
    "--invocable_version_handle",
    "-v",
    type=str,
    default=None,
    help="Handle of the package or plugin version being hosted.",
)
@click.option(
    "--invocable_instance_handle",
    "-h",
    type=str,
    default=None,
    help="Handle of the package or plugin instance being hosted.",
)
@click.option(
    "--api_key",
    "-k",
    type=str,
    default=None,
    help="API Key to hard-code for hosting.",
)
def serve(
    port: int = 8080,
    invocable_handle: Optional[str] = None,
    invocable_version_handle: Optional[str] = None,
    invocable_instance_handle: Optional[str] = None,
    api_key: Optional[str] = None,
):
    """Serve the local invocable"""
    initialize()
    path = find_api_py()
    api_module = get_api_module(path)
    invocable_class = get_class_from_module(api_module)
    click.secho(f"Found Invocable: {invocable_class.__name__}")

    server = SteamshipHTTPServer(
        invocable_class,
        port=port,
        invocable_handle=invocable_handle,
        invocable_version_handle=invocable_version_handle,
        invocable_instance_handle=invocable_instance_handle,
        default_api_key=api_key,
    )

    def on_exit(signum, frame):
        click.secho("Shutting down server.")
        server.stop()
        click.secho("Shut down.")
        exit(1)

    signal.signal(signal.SIGINT, on_exit)

    click.secho(f"Starting development server on port {server.port}")
    server.start()
    click.secho()


@click.command()
def deploy():
    """Deploy the package or plugin in this directory"""
    initialize()
    client = None
    try:
        client = Steamship()
    except SteamshipError as e:
        click.secho(e.message, fg="red")
        click.get_current_context().abort()

    user = User.current(client)
    if path.exists("steamship.json"):
        manifest = Manifest.load_manifest()
    else:
        manifest = manifest_init_wizard(client)
        manifest.save()

    if not path.exists("requirements.txt"):
        requirements_init_wizard()

    deployable_type = manifest.type

    update_config_template(manifest)

    deployer = None
    if deployable_type == DeployableType.PACKAGE:
        deployer = PackageDeployer()
    elif deployable_type == DeployableType.PLUGIN:
        deployer = PluginDeployer()
    else:
        click.secho("Deployable must be of type package or plugin.", fg="red")
        click.get_current_context().abort()

    deployable = deployer.create_or_fetch_deployable(client, user, manifest)

    click.echo("Bundling content... ", nl=False)
    bundle_deployable(manifest)
    click.echo("Done. 📦")

    _ = deployer.create_version(client, manifest, deployable.id)

    thing_url = f"{client.config.web_base}{deployable_type}s/{manifest.handle}"
    click.echo(
        f"Deployment was successful. View and share your new {deployable_type} here:\n\n{thing_url}\n"
    )

    # Common error conditions:
    # - Package/plugin handle already taken. [handled; asks user for new]
    # - Version handle already deployed. [handled; asks user for new]
    # - Bad parameter configuration. [mitigated by deriving template from Config object]
    # - Package content fails health check (ex. bad import) [Error caught while checking config object]


@click.command()
def info():
    """Displays information about the current Steamship user.

    This is useful to help users (in a support or hackathon context) test whether they have configured their
    Steamship environment correctly.
    """
    initialize()
    click.echo("\nSteamship Client Info\n=====================\n")

    if Configuration.default_config_file_has_api_key() or getenv("STEAMSHIP_API_KEY", None):
        # User is logged in!
        client = None
        try:
            client = Steamship()
        except BaseException:
            click.secho("Incorrect API key or network error.\n", fg="red")
            click.secho(
                "Your Steamship API Key is set, but we were unable to use it to fetch your account information.\n"
            )
            click.secho("- If you are on your own computer, run `ship login` to login.")
            click.secho(
                "- If you are in Replit, add the STEAMSHIP_API_KEY secret, then close and re-open this shell.\n"
            )
            return

        try:
            user = User.current(client)
            click.echo(f"User handle: {user.handle}")
            click.echo(f"User ID:     {user.id}")
            click.echo(f"Profile:     {client.config.profile}")
            click.echo("\nReady to ship! 🚢🚢🚢\n")

        except BaseException:
            click.secho("Incorrect API key or network error.\n", fg="red")
            click.secho(
                "Your Steamship API Key is set, but we were unable to use it to fetch your account information.\n"
            )
            click.secho("- If you are on your own computer, run `ship login` to login.")
            click.secho(
                "- If you are in Replit, add the STEAMSHIP_API_KEY secret, then close and re-open this shell.\n"
            )
    else:
        click.secho("You are not logged in.\n")
        click.secho("- If you are on your own computer, run `ship login` to login.")
        click.secho(
            "- If you are in Replit, add the STEAMSHIP_API_KEY secret, then close and re-open this shell.\n"
        )


@click.command()
@click.option(
    "--workspace",
    "-w",
    required=True,
    type=str,
    help="Workspace handle used for scoping logs request. All requests MUST be scoped by workspace.",
)
@click.option(
    "--offset",
    "-o",
    default=0,
    type=int,
    help="Starting index from sorted logs to return a chunk. Used for paging. Defaults to 0.",
)
@click.option(
    "--number",
    "-n",
    default=50,
    type=int,
    help="Number of logs to return in a single batch. Defaults to 50.",
)
@click.option(
    "--package",
    "-p",
    type=str,
    help="Package handle. Used to filter logs returend to a specific package (across all instances).",
)
@click.option(
    "--instance",
    "-i",
    type=str,
    help="Instance handle. Used to filter logs returned to a specific instance of a package.",
)
@click.option(
    "--version",
    "-v",
    type=str,
    help="Version handle. Used to filter logs returned to a specific version of a package.",
)
@click.option(
    "--path",
    "request_path",
    type=str,
    help="Path invoked by a client operation. Used to filter logs returned to a specific invocation path.",
)
def logs(
    workspace: str,
    offset: int,
    number: int,
    package: Optional[str] = None,
    instance: Optional[str] = None,
    version: Optional[str] = None,
    request_path: Optional[str] = None,
):
    initialize(suppress_message=True)
    client = None
    try:
        client = Steamship(workspace=workspace)
    except SteamshipError as e:
        raise click.UsageError(message=e.message)

    click.echo(json.dumps(client.logs(offset, number, package, instance, version, request_path)))


cli.add_command(login)
cli.add_command(deploy)
cli.add_command(info)
cli.add_command(deploy, name="it")
cli.add_command(ships)
cli.add_command(logs)
cli.add_command(serve)
cli.add_command(create_instance, name="use")


if __name__ == "__main__":
    serve([])



from typing import Optional

from assets.packages.configurable_hello_world import HelloWorld
from pydantic import Field

from steamship.invocable.config import Config, ConfigParameterType


def test_config_parameters():
    config_template = HelloWorld.get_config_parameters()
    assert len(config_template) == 4
    for config_param in config_template.values():
        assert config_param.default is None
        assert config_param.type == ConfigParameterType.STRING

    names = config_template.keys()
    assert "greeting" in names
    assert "snake_case_config" in names
    assert "camelCaseConfig" in names
    assert "defaultConfig" in names


def test_config_parameter_numbers():
    class NumberConfig(Config):
        x: float
        y: int

    config_template = NumberConfig.get_config_parameters()
    assert len(config_template) == 2
    for config_param in config_template.values():
        assert config_param.default is None
        assert config_param.type == ConfigParameterType.NUMBER

    names = config_template.keys()
    assert "x" in names
    assert "y" in names


def test_config_parameter_boolean():
    class BoolConfig(Config):
        x: bool
        y: bool

    bool_config = BoolConfig(x=True, y=False)
    config_template = bool_config.get_config_parameters()
    assert len(config_template) == 2
    for config_param in config_template.values():
        assert config_param.default is None
        assert config_param.type == ConfigParameterType.BOOLEAN

    names = config_template.keys()
    assert "x" in names
    assert "y" in names


def test_config_parameter_with_defaults():
    class DefaultsConfig(Config):
        w: str = "deeeefault"
        x: bool = True
        y: int = 3
        z: float = 7.5

    defaults_config = DefaultsConfig()
    config_template = defaults_config.get_config_parameters()
    assert len(config_template) == 4

    assert config_template["w"].type == ConfigParameterType.STRING
    assert config_template["w"].default == "deeeefault"
    assert config_template["x"].type == ConfigParameterType.BOOLEAN
    assert config_template["x"].default  # Assert that the default == True
    assert config_template["y"].type == ConfigParameterType.NUMBER
    assert config_template["y"].default == 3
    assert config_template["z"].type == ConfigParameterType.NUMBER
    assert config_template["z"].default == 7.5


def test_optional_config_parameters():
    class DefaultsConfig(Config):
        w: Optional[str] = "deeeefault"
        w2: Optional[str]
        x: Optional[bool] = True
        x2: Optional[bool]
        y: Optional[int] = 3
        y2: Optional[int]
        z: Optional[float] = 7.5
        z2: Optional[float]

    defaults_config = DefaultsConfig()
    config_template = defaults_config.get_config_parameters()
    assert len(config_template) == 8

    assert config_template["w"].type == ConfigParameterType.STRING
    assert config_template["w"].default == "deeeefault"
    assert config_template["x"].type == ConfigParameterType.BOOLEAN
    assert config_template["x"].default  # Assert that the default == True
    assert config_template["y"].type == ConfigParameterType.NUMBER
    assert config_template["y"].default == 3
    assert config_template["z"].type == ConfigParameterType.NUMBER
    assert config_template["z"].default == 7.5

    assert config_template["w2"].type == ConfigParameterType.STRING
    assert config_template["w2"].default is None
    assert config_template["x2"].type == ConfigParameterType.BOOLEAN
    assert config_template["x2"].default is None
    assert config_template["y2"].type == ConfigParameterType.NUMBER
    assert config_template["y2"].default is None
    assert config_template["z2"].type == ConfigParameterType.NUMBER
    assert config_template["z2"].default is None


def test_descriptions():
    class DescriptionConfig(Config):
        x: int = Field(description="Test Description")

    description_config = DescriptionConfig(x=3)
    config_template = description_config.get_config_parameters()
    assert len(config_template) == 1

    # TODO: descriptions
    assert config_template["x"].description == "Test Description"



import json
import logging
from pathlib import Path
from typing import List, Optional, Type

from steamship import File, SteamshipError, Tag
from steamship.base import Task, TaskState
from steamship.invocable import Config, InvocableResponse, create_handler
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput
from steamship.plugin.outputs.training_parameter_plugin_output import TrainingParameterPluginOutput
from steamship.plugin.request import PluginRequest
from steamship.plugin.tagger import TrainableTagger
from steamship.plugin.trainable_model import TrainableModel

TRAINING_PARAMETERS = TrainingParameterPluginOutput(
    training_epochs=3,
    testing_holdout_percent=0.3,
    training_params={"keywords": ["chocolate", "roses", "champagne"]},
)


def make_train_response():
    return InvocableResponse(data=TrainPluginOutput())


class EmptyConfig(Config):
    pass


class TestTrainableTaggerModel(TrainableModel[EmptyConfig]):
    """Example of a trainable model.

    At some point, may want to evolve this into an abstract base class, but for the time being, here is the
    design of what a "Model" ought to look like:

    TRAINING:

        - Save all results to the folder provided during initialization. The zipfile of this folder
            is the ModelCheckpoint that gets saved to steamship.
        - Note that we'll likely want to iterate on this, e.g. Does the model create the checkpoint or simply receive a
            folder into which to save data, etc etc.


    LOADING:
    - Offer a constructor that loads the model with a provided path. This path is assumed to contain the unzipped
    ModelCheckpoint that was requested.

      In this example model, the `from_disk(path: Path)` method provides this functionality.
    """

    KEYWORD_LIST_FILE = (
        "keyword_list.json"  # File, relative to the checkpoint, to save/load features from.
    )
    keyword_list: List[str] = None  # Contents of the KEYWORD_LIST_FILE file
    path: Optional[Path] = None  # Save the effective path for testing

    def __init__(self):
        self.path = None
        self.keyword_list = []

    def load_from_folder(self, checkpoint_path: Path):
        """[Required by TrainableModel] Load state from the provided path."""
        logging.info(f"Model:save_to_folder {checkpoint_path}")
        self.path = checkpoint_path
        with open(self.path / TestTrainableTaggerModel.KEYWORD_LIST_FILE, "r") as f:
            self.keyword_list = json.loads(f.read())

    def save_to_folder(self, checkpoint_path: Path):
        """[Required by TrainableModel] Save state to the provided path."""
        logging.info(f"Model:save_to_folder {checkpoint_path}")
        self.path = checkpoint_path
        with open(checkpoint_path / TestTrainableTaggerModel.KEYWORD_LIST_FILE, "w") as f:
            f.write(json.dumps(self.keyword_list))

    def train(self, input: PluginRequest[TrainPluginInput]) -> InvocableResponse[TrainPluginOutput]:
        """Training for this model is to set the parameters to those provided in the input object.

        This allows us to test that we're properly passing through the training parameters to the train process.

        This training happens synchronously -- the train_status call should never happen!
        """
        logging.info("TestTrainableTaggerModel:train()")
        self.keyword_list = input.data.training_params.get("keyword_list", [])
        return make_train_response()

    def train_status(
        self, input: PluginRequest[TrainPluginInput]
    ) -> InvocableResponse[TrainPluginOutput]:
        """Training for this model is to set the parameters to those provided in the input object.

        This allows us to test that we're properly passing through the training parameters to the train process.
        """
        raise SteamshipError(
            message="The train_status method should never be called on this tagger!"
        )

    def run(
        self, request: PluginRequest[BlockAndTagPluginInput]
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        """Tags the incoming data for any instance of the keywords in the parameter file."""
        logging.info(f"TestTrainableTaggerModel:run() - My keyword list is {self.keyword_list}")
        response = InvocableResponse(
            data=BlockAndTagPluginOutput(
                file=File(tags=[Tag(kind=word) for word in self.keyword_list])
            )
        )
        logging.info(f"TestTrainableTaggerModel:run() returning {response}")
        return response


class TestTrainableTaggerPlugin(TrainableTagger):
    """Tests the Trainable Tagger lifecycle.

    - This tagger produces a FIXED set of trainable parameters.
    - These parameters (and not the trainable data!) fully parameterize the trained model.
    - The trained model (and not the trainable parameters!) fully parameterize the running model.
    - The model simply tags keywords that it finds in the text.

    Taken together, this plugin can be seen as a reference implementation of the data/process lifecycle of a trainable
    model, regardless of where the actual work occurs:

    - It could occur here, running in Lambda.
    - It could occur here, running in ECS.
    - It could be orchestrated from here, but runs in HuggingFace / SageMaker / or elsewhere

    """

    @classmethod
    def config_cls(cls) -> Type[Config]:
        return EmptyConfig

    def model_cls(self) -> Type[TestTrainableTaggerModel]:
        return TestTrainableTaggerModel

    def run_with_model(
        self,
        request: PluginRequest[BlockAndTagPluginInput],
        model: TestTrainableTaggerModel,
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        """Downloads the model file from the provided workspace"""
        logging.debug(f"run_with_model {request} {model}")
        logging.info(
            f"TestTrainableTaggerPlugin:run_with_model() got request {request} and model {model}"
        )
        return model.run(request)

    def get_training_parameters(
        self, request: PluginRequest[TrainingParameterPluginInput]
    ) -> InvocableResponse[TrainingParameterPluginOutput]:
        ret = InvocableResponse[TrainingParameterPluginOutput](data=TRAINING_PARAMETERS)
        return ret

    def train(
        self, request: PluginRequest[TrainPluginInput], model: TestTrainableTaggerModel
    ) -> InvocableResponse[TrainPluginOutput]:
        """Since trainable can't be assumed to be asynchronous, the trainer is responsible for uploading its own model file."""
        logging.info(f"TestTrainableTaggerPlugin:train() {request}")

        # Create a Response object at the top with a Task attached. This will let us stream back updates
        # TODO: This is very non-intuitive. We should improve this.
        response = InvocableResponse(status=Task(state=TaskState.running))

        # Example of recording training progress
        # response.status.status_message = "About to train!"
        # response.post_update(client=self.client)

        # Train the model
        train_plugin_output = model.train(request)

        # Save the model with the `default` handle.
        archive_path_in_steamship = model.save_remote(
            client=self.client, plugin_instance_id=request.context.plugin_instance_id
        )

        # Set the model location on the plugin output.
        logging.info(
            f"TestTrainableTaggerPlugin:train() setting model archive path to {archive_path_in_steamship}"
        )
        train_plugin_output.data.archive_path = archive_path_in_steamship

        # Set the response on the `data` field of the object
        response.set_data(json=train_plugin_output.data)
        response.status.state = TaskState.succeeded
        response.status.status_message = "Done!"

        # If we want we can post this to the Engine
        # response.status.status_message = "Done!"
        # response.status.state = TaskState.succeeded
        # response.post_update(client=self.client)

        # Or, if this training really did happen synchronously, we return it.
        logging.info(
            f"TestTrainableTaggerPlugin:train() returning ###{json.dumps(response.dict(by_alias=True))}###--"
        )
        return response

    def train_status(
        self, request: PluginRequest[TrainPluginInput], model: TrainableModel
    ) -> InvocableResponse[TrainPluginOutput]:
        # This plugin never keeps a training task going beyond one function call.  This method should not be called.
        raise SteamshipError(message="The train_status call should not happen on this model.")


handler = create_handler(TestTrainableTaggerPlugin)



from .action import Action, FinishAction
from .agent import Agent, LLMAgent
from .context import AgentContext, EmitFunc, Metadata
from .llm import LLM
from .output_parser import OutputParser
from .tool import Tool

__all__ = [
    "Action",
    "Agent",
    "AgentContext",
    "EmitFunc",
    "FinishAction",
    "Metadata",
    "LLM",
    "LLMAgent",
    "OutputParser",
    "Tool",
]



from typing import Any, Dict, List, Optional, Union, cast

from pydantic import Field

from steamship.base.error import SteamshipError
from steamship.base.model import CamelModel
from steamship.base.tasks import Task
from steamship.data.embeddings import EmbeddedItem, EmbeddingIndex, QueryResult, QueryResults
from steamship.data.plugin.plugin_instance import PluginInstance
from steamship.data.tags.tag import Tag


class EmbedderInvocation(CamelModel):
    """The parameters capable of creating/fetching an Embedder (Tagger) Plugin Instance."""

    plugin_handle: str
    instance_handle: Optional[str] = None
    config: Optional[Dict[str, Any]] = None
    version: Optional[str] = None
    fetch_if_exists: bool = True


class SearchResult(CamelModel):
    """A single scored search result -- which is always a tag.

    This class is intended to eventually replace the QueryResult object currently used with the Embedding layer."""

    tag: Optional[Tag] = None
    score: Optional[float] = None

    @staticmethod
    def from_query_result(query_result: QueryResult) -> "SearchResult":
        hit = query_result.value
        value = hit.metadata or {}

        # To make this change Python-only, some fields are stached in `hit.metadata`.
        # This has the temporary consequence of these keys not being safe. This will be resolved when we spread
        # this refactor to the engine.
        block_id = None
        if "_block_id" in value:
            block_id = value.get("_block_id")
        del value["_block_id"]

        file_id = None
        if "_file_id" in value:
            file_id = value.get("_file_id")
        del value["_file_id"]

        tag_id = None
        if "_tag_id" in value:
            tag_id = value.get("_tag_id")
        del value["_tag_id"]

        tag = Tag(
            id=hit.id,
            kind=hit.external_type,
            name=hit.external_id,
            block_id=block_id,
            tag_id=tag_id,
            file_id=file_id,
            text=hit.value,
            value=value,
        )
        return SearchResult(tag=tag, score=query_result.score)


class SearchResults(CamelModel):
    """Results of a search operation -- which is always a list of ranked tag.

    This class is intended to eventually replace the QueryResults object currently used with the Embedding layer.
    TODO: add in paging support."""

    items: List[SearchResult] = None

    @staticmethod
    def from_query_results(query_results: QueryResults) -> "SearchResults":
        items = [SearchResult.from_query_result(qr) for qr in query_results.items or []]
        return SearchResults(items=items)


class EmbeddingIndexPluginInstance(PluginInstance):
    """A persistent, read-optimized index over embeddings.

    This is currently implemented as an object which behaves like a PluginInstance even though
    it isn't from an implementation perspective on the back-end.
    """

    embedder: PluginInstance = Field(None, exclude=True)
    index: EmbeddingIndex = Field(None, exclude=True)

    def reset(self):
        self.index.delete()
        self.index = EmbeddingIndex.create(
            client=self.client,
            handle=self.handle,
            embedder_plugin_instance_handle=self.embedder.handle,
            fetch_if_exists=False,
        )

    def delete(self):
        """Delete the EmbeddingIndexPluginInstnace.

        For now, we will have this correspond to deleting the `index` but not the `embedder`. This is likely
        a temporary design.
        """
        return self.index.delete()

    def insert(self, tags: Union[Tag, List[Tag]], allow_long_records: bool = False):
        """Insert tags into the embedding index."""

        # Make a list if a single tag was provided
        if isinstance(tags, Tag):
            tags = [tags]

        for tag in tags:
            if not tag.text:
                raise SteamshipError(
                    message="Please set the `text` field of your Tag before inserting it into an index."
                )

            # Now we need to prepare an EmbeddingIndexItem of a particular shape that encodes the tag.
            metadata = tag.value or {}
            if not isinstance(metadata, dict):
                raise SteamshipError(
                    "Only Tags with a dict or None value can be embedded. "
                    + f"This tag had a value of type: {type(tag.value)}"
                )

            # To make this change Python-only, some fields are stached in `hit.metadata`.
            # This has the temporary consequence of these keys not being safe. This will be resolved when we spread
            # this refactor to the engine.
            metadata["_file_id"] = tag.file_id
            metadata["_tag_id"] = tag.id
            metadata["_block_id"] = tag.block_id
            tag.value = metadata

        embedded_items = [
            EmbeddedItem(
                value=tag.text,
                external_id=tag.name,
                external_type=tag.kind,
                metadata=tag.value,
            )
            for tag in tags
        ]

        # We always reindex in this new style; to not do so is to expose details (when embedding occurrs) we'd rather
        # not have users exercise control over.
        self.index.insert_many(embedded_items, reindex=True, allow_long_records=allow_long_records)

    def search(self, query: str, k: Optional[int] = None) -> Task[SearchResults]:
        """Search the embedding index.

        This wrapper implementation simply projects the `Hit` data structure into a `Tag`
        """
        if query is None or len(query.strip()) == 0:
            raise SteamshipError(message="Query field must be non-empty.")

        # Metadata will always be included; this is the equivalent of Tag.value
        wrapped_result = self.index.search(query, k=k, include_metadata=True)

        # For now, we'll have to do this synchronously since we're trying to avoid changing things on the engine.
        wrapped_result.wait()

        # We're going to do a switcheroo on the output type of Task here.
        search_results = SearchResults.from_query_results(wrapped_result.output)
        wrapped_result.output = search_results

        # Return the index's search result, but projected into the data structure of Tags
        return cast(Task[SearchResults], wrapped_result)

    @staticmethod
    def create(
        client: Any,
        plugin_id: str = None,
        plugin_handle: str = None,
        plugin_version_id: str = None,
        plugin_version_handle: str = None,
        handle: str = None,
        fetch_if_exists: bool = True,
        config: Dict[str, Any] = None,
    ) -> "EmbeddingIndexPluginInstance":
        """Create a class that simulates an embedding index re-implemented as a PluginInstance."""

        # Perform a manual config validation check since the configuration isn't actually being sent up to the Engine.
        # In this case, an embedding index has special behavior which is to instantiate/fetch an Embedder that it can use.
        if "embedder" not in config:
            raise SteamshipError(
                message="Config key missing. Please include a field named `embedder` with type `EmbedderInvocation`."
            )

        # Just for pydantic validation.
        embedder_invocation = EmbedderInvocation.parse_obj(config["embedder"])

        # Create the embedder
        embedder = client.use_plugin(**embedder_invocation.dict())
        embedder.wait_for_init()

        # Create the index
        index = EmbeddingIndex.create(
            client=client,
            handle=handle,
            embedder_plugin_instance_handle=embedder.handle,
            fetch_if_exists=fetch_if_exists,
        )

        # Now return the plugin wrapper
        return EmbeddingIndexPluginInstance(
            client=client, id=index.id, handle=index.handle, index=index, embedder=embedder
        )



from steamship.agents.tools.text_generation.text_rewrite_tool import TextRewritingTool
from steamship.utils.repl import ToolREPL

DEFAULT_PROMPT = """Instructions:
Please rewrite the following passage to create an excellent prompt for use with DALL-E or Stable Diffusion. Add
keywords that improve the drama and style to create dramatic, photographic effects.

Passage:
{input}

Image Generation Prompt:"""


class ImagePromptGeneratorTool(TextRewritingTool):
    """
    Example tool to illustrate rewriting an input query to become a better prompt.
    """

    name: str = "ImagePromptGenerator"
    human_description: str = "Improves a prompt for use with image generation."
    agent_description: str = (
        "Use this tool to improve a prompt for stable diffusion and other image and video generators. "
        "This tool will refine your prompt to include key words and phrases that make "
        "stable diffusion and other art generation algorithms perform better. The input is a prompt text string "
        "and the output is a prompt text string"
    )
    rewrite_prompt: str = DEFAULT_PROMPT


if __name__ == "__main__":
    ToolREPL(ImagePromptGeneratorTool()).run()



from typing import Optional

from steamship.agents.schema import AgentContext
from steamship.agents.schema.llm import LLM

_LLM_KEY = "llm"


def with_llm(llm: LLM, context: Optional[AgentContext] = None) -> AgentContext:
    """Sets an LLM for general purpose lookup and usage on an AgentContext."""
    if context is None:
        # TODO: should we have a default context somehow?
        context = AgentContext()
        context.completed_steps = []
    context.metadata[_LLM_KEY] = llm
    return context


def get_llm(context: AgentContext) -> Optional[LLM]:
    """Retrieves the LLM from the provided AgentContext (if it exists)."""
    return context.metadata.get(_LLM_KEY, None)



import json
import logging
import re
from http import server
from typing import Dict, Optional, Type
from urllib.parse import parse_qs, urlparse

from steamship import MimeTypes, Steamship
from steamship.data.user import User
from steamship.invocable import (
    Invocable,
    InvocableRequest,
    Invocation,
    InvocationContext,
    LoggingConfig,
    PackageService,
)
from steamship.invocable.lambda_handler import handler, internal_handler


def create_safe_handler(invocable: Type[Invocable] = None):
    """Mimics the create_safe_handler function in lambda_handler for parallelism."""
    invocable_getter = lambda: invocable  # noqa: E731
    bound_internal_handler = lambda event, client, context: internal_handler(  # noqa: E731
        invocable_getter, event, client, context, call_instance_init=True
    )
    return lambda event, context=None: handler(
        bound_internal_handler, event, context, running_locally=True
    )


def make_handler(  # noqa: C901
    invocable_class: Type[Invocable],  # The invocable (package or plugin) that this handler hosts.
    port: int,
    base_url: Optional[str] = "http://localhost",
    default_api_key: Optional[str] = None,
    invocable_handle: str = None,
    invocable_version_handle: str = None,
    invocable_instance_handle: str = None,
    config: dict = None,
):
    """Creates and returns a SimpleHTTPRequestHandler class for an Invocable (package or plugin).

    For use with steamship.cli.http.server.Server.
    """
    # A cache of API Key to User objects, to avoid doing a lookup each request
    user_for_key: Dict[str, User] = {}

    # The type of invocable this handler hosts
    invocable_type: Optional[str]

    if issubclass(invocable_class, PackageService):
        invocable_type = "package"
    else:
        invocable_type = "plugin"

    class InvocableHTTPHandler(server.SimpleHTTPRequestHandler):
        def _send_response(self, _bytes: bytes, mime_type: MimeTypes):
            self.send_response(200)
            self.send_header("Content-type", mime_type)
            self.end_headers()
            self.wfile.write(_bytes)

        def _get_client(self) -> Steamship:
            """Returns a Steamship client.

            The API Key used in the client will be chosen in this order:
            - From the bearer token, if provided,
            - From self.default_api_key, if provided
            - From ~/.steamship.json, if present

            Note that this means the broader system on which this server is running is capable of seeding this server
            with a default API Key. If you do not want this behavior, do not keep a .steamship.json file in your home
            directory.
            """
            api_key = default_api_key
            for h in self.headers:
                if h.lower() == "authorization":
                    api_key = re.sub(r"^[Bb]earer\s+", "", self.headers[h])
            return Steamship(api_key=api_key)

        def _get_invocation_context(self, client) -> InvocationContext:
            nonlocal user_for_key

            # Get the cached user or fetch a new one
            if client.config.api_key in user_for_key:
                user = user_for_key[client.config.api_key]
            else:
                user = User.current(client)
                user_for_key[client.config.api_key] = user

            url = f"{base_url}:{port}/"

            return InvocationContext(
                user_id=user.id,
                workspace_id=client.config.workspace_id,
                workspace_handle=client.config.workspace_handle,
                invocable_owner_id=user.id,
                invocable_owner_handle=user.handle,
                invocable_handle=invocable_handle,
                invocable_version_handle=invocable_version_handle,
                invocable_instance_handle=invocable_instance_handle,
                invocable_type=invocable_type,
                invocable_url=url,
            )

        def _do_request(self, payload: dict, http_verb: str):
            try:
                client = self._get_client()
                context = self._get_invocation_context(client)

                invocation = Invocation(
                    http_verb=http_verb, invocation_path=self.path, arguments=payload, config=config
                )
                event = InvocableRequest(
                    client_config=client.config,
                    invocation=invocation,
                    logging_config=LoggingConfig(logging_host=None, logging_port=None),
                    invocation_context=context,
                )

                handler = create_safe_handler(invocable_class)
                resp = handler(event.dict(by_alias=True), context)
                res_str = json.dumps(resp)
                res_bytes = bytes(res_str, "utf-8")
                self._send_response(res_bytes, MimeTypes.JSON)
            except Exception as e:
                self._send_response(bytes(f"{e}", "utf-8"), MimeTypes.TXT)

        def do_GET(self):  # noqa: N802
            logging.info(
                "GET request,\nPath: %s\nHeaders:\n%s\n", str(self.path), str(self.headers)
            )
            try:
                payload = parse_qs(urlparse(self.path).query)
                return self._do_request(payload, "GET")
            except Exception as e:
                self._send_response(bytes(f"{e}", "utf-8"), MimeTypes.TXT)

        def do_POST(self):  # noqa: N802
            logging.info(
                "POST request,\nPath: %s\nHeaders:\n%s\n", str(self.path), str(self.headers)
            )
            content_length = int(self.headers["Content-Length"])  # <--- Gets the size of data
            post_data = self.rfile.read(content_length)  # <--- Gets the data itself
            try:
                data_str = post_data.decode("utf8")
                payload = json.loads(data_str)
                return self._do_request(payload, "POST")
            except Exception as e:
                self._send_response(bytes(f"{e}", "utf-8"), MimeTypes.TXT)

    return InvocableHTTPHandler
